from pathlib import Path

import numpy as np
import numpy.testing as npt
import pytest
import rasterio as rio
from osgeo import gdal

from dolphin.io._readers import (
    BinaryReader,
    BinaryStackReader,
    EagerLoader,
    HDF5Reader,
    HDF5StackReader,
    RasterReader,
    RasterStackReader,
    VRTStack,
    _parse_vrt_file,
)
from dolphin.utils import _get_path_from_gdal_str

# Note: uses the fixtures from conftest.py

# Get combinations of slices
slices_to_test = [slice(None), 1, slice(0, 10, 2)]


@pytest.fixture(scope="module")
def binary_file_list(tmp_path_factory, slc_stack):
    """Flat binary files in the ENVI format."""
    import rasterio as rio
    from rasterio.errors import NotGeoreferencedWarning

    shape = slc_stack[0].shape
    dtype = slc_stack.dtype
    tmp_path = tmp_path_factory.mktemp("data")

    # Create a stack of binary files
    files = []
    for i, slc in enumerate(slc_stack):
        f = tmp_path / f"test_{i}.bin"
        # Ignore warning
        with pytest.warns(NotGeoreferencedWarning), rio.open(
            f,
            "w",
            driver="ENVI",
            width=shape[1],
            height=shape[0],
            count=1,
            dtype=dtype,
        ) as dst:
            dst.write(slc, 1)
        files.append(f)

    return files


@pytest.fixture()
def binary_reader(slc_stack, binary_file_list):
    f = BinaryReader(
        binary_file_list[0], shape=slc_stack[0].shape, dtype=slc_stack.dtype
    )
    assert f.shape == slc_stack[0].shape
    assert f.dtype == slc_stack[0].dtype
    return f


class TestBinary:
    def test_binary_file_read(self, binary_reader, slc_stack):
        npt.assert_array_almost_equal(binary_reader[()], slc_stack[0])
        # Check the reading of a subset
        npt.assert_array_almost_equal(
            binary_reader[0:10, 0:10], slc_stack[0][0:10, 0:10]
        )

    @pytest.fixture(scope="module")
    def binary_stack(self, slc_stack, binary_file_list):
        s = BinaryStackReader.from_file_list(
            binary_file_list, shape_2d=slc_stack[0].shape, dtype=slc_stack.dtype
        )
        assert s.shape == slc_stack.shape
        assert len(s) == len(slc_stack) == len(binary_file_list)
        assert s.ndim == 3
        assert s.dtype == slc_stack.dtype
        return s

    @pytest.mark.parametrize("dslice", slices_to_test)
    @pytest.mark.parametrize("rslice", slices_to_test)
    @pytest.mark.parametrize("cslice", slices_to_test)
    def test_binary_stack_read_slices(
        self, binary_stack, slc_stack, dslice, rslice, cslice
    ):
        s = binary_stack[dslice, rslice, cslice]
        expected = slc_stack[dslice, rslice, cslice]
        assert s.shape == expected.shape
        npt.assert_array_almost_equal(s, expected)


# #### HDF5 Tests ####


@pytest.fixture(scope="module")
def hdf5_file_list(tmp_path_factory, slc_stack):
    """Flat binary files in the ENVI format."""
    import h5py

    tmp_path = tmp_path_factory.mktemp("data")

    # Create a stack of binary files
    files = []
    for i, slc in enumerate(slc_stack):
        f = tmp_path / f"test_{i}.h5"
        with h5py.File(f, "w") as dst:
            dst.create_dataset("data", data=slc)
        files.append(f)

    return files


@pytest.fixture()
def hdf5_reader(hdf5_file_list, slc_stack):
    r = HDF5Reader(hdf5_file_list[0], dset_name="data", keep_open=True)
    assert r.shape == slc_stack[0].shape
    assert r.dtype == slc_stack[0].dtype
    return r


class TestHDF5:
    def test_hdf5_reader_read(self, hdf5_reader, slc_stack):
        npt.assert_array_almost_equal(hdf5_reader[()], slc_stack[0])
        # Check the reading of a subset
        npt.assert_array_almost_equal(hdf5_reader[0:10, 0:10], slc_stack[0][0:10, 0:10])

    @pytest.mark.parametrize("keep_open", [True, False])
    def hdf5_stack(self, hdf5_file_list, slc_stack, keep_open):
        s = HDF5StackReader.from_file_list(
            hdf5_file_list, dset_names="data", keep_open=keep_open
        )
        assert s.shape == slc_stack.shape
        assert len(s) == len(slc_stack) == len(hdf5_file_list)
        assert s.ndim == 3
        assert s.dtype == slc_stack.dtype
        return s

    @pytest.mark.parametrize("dslice", slices_to_test)
    @pytest.mark.parametrize("rslice", slices_to_test)
    @pytest.mark.parametrize("cslice", slices_to_test)
    @pytest.mark.parametrize("keep_open", [True, False])
    def test_hdf5_stack_read_slices(
        self, hdf5_file_list, slc_stack, keep_open, dslice, rslice, cslice
    ):
        reader = HDF5StackReader.from_file_list(
            hdf5_file_list, dset_names="data", keep_open=keep_open
        )
        s = reader[dslice, rslice, cslice]
        expected = slc_stack[dslice, rslice, cslice]
        assert s.shape == expected.shape
        npt.assert_array_almost_equal(s, expected)


# #### RasterReader Tests ####
@pytest.fixture()
def raster_reader(slc_file_list, slc_stack):
    # ignore georeferencing warnings
    with pytest.warns(rio.errors.NotGeoreferencedWarning):
        r = RasterReader.from_file(slc_file_list[0])
    assert r.shape == slc_stack[0].shape
    assert r.dtype == slc_stack[0].dtype
    assert r.ndim == 2
    assert r.dtype == np.complex64
    return r


class TestRaster:
    @pytest.mark.parametrize("keep_open", [True, False])
    def test_raster_stack_reader(
        self,
        slc_file_list,
        slc_stack,
        keep_open,
    ):
        with pytest.warns(rio.errors.NotGeoreferencedWarning):
            reader = RasterStackReader.from_file_list(
                slc_file_list, keep_open=keep_open
            )
            assert reader.ndim == 3
            assert reader.shape == slc_stack.shape
            assert reader.dtype == slc_stack.dtype
            assert len(reader) == len(slc_stack) == len(slc_file_list)

    @pytest.mark.parametrize("dslice", slices_to_test)
    @pytest.mark.parametrize("rslice", slices_to_test)
    @pytest.mark.parametrize("cslice", slices_to_test)
    @pytest.mark.parametrize("keep_open", [True, False])
    def test_raster_stack_read_slices(
        self, slc_file_list, slc_stack, keep_open, dslice, rslice, cslice
    ):
        with pytest.warns(rio.errors.NotGeoreferencedWarning):
            reader = RasterStackReader.from_file_list(
                slc_file_list, keep_open=keep_open
            )
            s = reader[dslice, rslice, cslice]
        expected = slc_stack[dslice, rslice, cslice]
        assert s.shape == expected.shape
        npt.assert_array_almost_equal(s, expected)


@pytest.mark.parametrize("rows", slices_to_test)
@pytest.mark.parametrize("cols", slices_to_test)
def test_ellipsis_reads(binary_reader, hdf5_reader, raster_reader, rows, cols):
    # Test that the ellipsis works
    npt.assert_array_equal(binary_reader[...], binary_reader[()])
    npt.assert_array_equal(hdf5_reader[...], hdf5_reader[()])
    npt.assert_array_equal(raster_reader[...], raster_reader[()])
    # Test we can still do rows/cols with a leading ellipsis
    npt.assert_array_equal(binary_reader[..., rows, cols], binary_reader[rows, cols])
    npt.assert_array_equal(hdf5_reader[..., rows, cols], hdf5_reader[rows, cols])
    npt.assert_array_equal(raster_reader[..., rows, cols], raster_reader[rows, cols])


# #### VRT Tests ####


@pytest.fixture()
def vrt_stack(tmp_path, slc_stack, slc_file_list):
    vrt_file = tmp_path / "test.vrt"
    s = VRTStack(slc_file_list, outfile=vrt_file)

    assert s.shape == slc_stack.shape
    assert len(s) == len(slc_stack) == len(slc_file_list)
    return s


@pytest.fixture()
def vrt_stack_nc(tmp_path, slc_stack, slc_file_list_nc):
    vrt_file = tmp_path / "test_nc.vrt"
    s = VRTStack(slc_file_list_nc, outfile=vrt_file, subdataset="data")

    assert s.shape == slc_stack.shape
    return s


@pytest.fixture()
def vrt_stack_nc_subdataset(tmp_path, slc_stack, slc_file_list_nc_with_sds):
    vrt_file = tmp_path / "test_nc.vrt"
    files_only = [_get_path_from_gdal_str(f) for f in slc_file_list_nc_with_sds]
    s = VRTStack(files_only, outfile=vrt_file, subdataset="data/VV")

    assert s.shape == slc_stack.shape
    return s


@pytest.fixture()
def vrt_stack_nc_wgs84(tmp_path, slc_stack, slc_file_list_nc_wgs84):
    # Check an alternative projection system
    vrt_file = tmp_path / "test_nc_wgs84.vrt"
    s = VRTStack(slc_file_list_nc_wgs84, outfile=vrt_file)

    assert s.shape == slc_stack.shape
    return s


def test_create(vrt_stack, vrt_stack_nc):
    for _v in [vrt_stack, vrt_stack_nc]:
        vrt_file = vrt_stack.outfile
        assert vrt_file.exists()
        assert vrt_file.stat().st_size > 0

        # Check that the VRT is valid
        ds = gdal.Open(str(vrt_file))
        assert ds is not None
        ds = None


def test_create_over_existing(tmp_path, slc_file_list):
    vrt_file = tmp_path / "test.vrt"
    VRTStack(slc_file_list, outfile=vrt_file)
    VRTStack(slc_file_list, outfile=vrt_file, fail_on_overwrite=False)
    with pytest.raises(FileExistsError):
        VRTStack(slc_file_list, outfile=vrt_file, fail_on_overwrite=True)


def test_from_vrt_file(tmp_path, slc_file_list):
    vrt_file = tmp_path / "test.vrt"
    s = VRTStack(slc_file_list, outfile=vrt_file)
    s2 = VRTStack.from_vrt_file(vrt_file)
    assert s == s2


def test_read_stack(vrt_stack, slc_stack):
    ds = gdal.Open(str(vrt_stack.outfile))
    loaded = ds.ReadAsArray()
    npt.assert_array_almost_equal(loaded, slc_stack)

    data = vrt_stack.read_stack()
    npt.assert_array_almost_equal(data, slc_stack)
    assert data.ndim == vrt_stack.ndim
    assert data.shape == vrt_stack.shape


def test_read_stack_nc(vrt_stack_nc, slc_stack):
    ds = gdal.Open(str(vrt_stack_nc.outfile))
    loaded = ds.ReadAsArray()
    npt.assert_array_almost_equal(loaded, slc_stack)
    npt.assert_array_almost_equal(vrt_stack_nc.read_stack(), slc_stack)


def test_read_stack_1_file(tmp_path, slc_file_list):
    vrt_file = tmp_path / "test_1file.vrt"
    v = VRTStack(slc_file_list[:1], outfile=vrt_file)
    data = v.read_stack()
    assert data.ndim == v.ndim
    assert data.shape == v.shape

    data = v[:, :, :]
    assert data.ndim == v.ndim
    assert data.shape == v.shape


def test_sort_order(tmp_path, slc_file_list):
    random_order = [Path(f) for f in np.random.permutation(slc_file_list)]
    # Make sure the files are sorted by date
    vrt_stack = VRTStack(random_order, outfile=tmp_path / "test.vrt")
    assert vrt_stack.file_list == [Path(f) for f in slc_file_list]

    vrt_stack2 = VRTStack(
        random_order, sort_files=False, outfile=tmp_path / "test2.vrt"
    )
    assert vrt_stack2.file_list == random_order


def test_dates(vrt_stack, vrt_stack_nc_subdataset):
    dates = vrt_stack.dates
    assert len(dates) == len(vrt_stack)
    d0 = 20220101
    for d in dates:
        assert d[0].strftime("%Y%m%d") == str(d0)
        d0 += 1

    d0 = 20220101
    for d in vrt_stack_nc_subdataset.dates:
        assert d[0].strftime("%Y%m%d") == str(d0)
        d0 += 1


def test_bad_sizes(slc_file_list, raster_10_by_20):
    from dolphin.io import get_raster_xysize

    # Make sure the files are the same size
    assert get_raster_xysize(slc_file_list[0]) == get_raster_xysize(slc_file_list[1])
    assert get_raster_xysize(slc_file_list[0]) != get_raster_xysize(raster_10_by_20)
    with pytest.raises(ValueError):
        VRTStack([*slc_file_list, raster_10_by_20], outfile="other.vrt")


def test_iter_blocks(vrt_stack):
    loader = EagerLoader(reader=vrt_stack, block_shape=(5, 5))
    blocks, slices = zip(*list(loader.iter_blocks()))
    # (5, 10) total shape, breaks into 5x5 blocks
    assert len(blocks) == 2
    for b in blocks:
        assert b.shape == (len(vrt_stack), 5, 5)

    loader = EagerLoader(reader=vrt_stack, block_shape=(5, 2))
    blocks, slices = zip(*list(loader.iter_blocks()))
    assert len(blocks) == 5
    for b in blocks:
        assert b.shape == (len(vrt_stack), 5, 2)


def test_tiled_iter_blocks(tmp_path, tiled_file_list):
    outfile = tmp_path / "stack.vrt"
    vrt_stack = VRTStack(tiled_file_list, outfile=outfile)
    loader = EagerLoader(reader=vrt_stack, block_shape=(32, 32))
    blocks, slices = zip(*list(loader.iter_blocks()))
    # (100, 200) total shape, breaks into 32x32 blocks
    assert len(blocks) == len(slices) == 28
    for i, b in enumerate(blocks, start=1):
        # Account for the smaller block sizes at the ends
        if i % 7 == 0:
            # last col
            if i > 21:  # Last row
                assert b.shape == (len(vrt_stack), 4, 8)
            else:
                assert b.shape == (len(vrt_stack), 32, 8)

    loader = EagerLoader(reader=vrt_stack, block_shape=(50, 100))
    blocks, slices = zip(*list(loader.iter_blocks()))
    assert len(blocks) == len(slices) == 4


@pytest.fixture()
def test_vrt():
    return """<VRTDataset rasterXSize="128" rasterYSize="128">
  <SRS dataAxisToSRSAxisMapping="1,2">PROJCS["WGS 84 / UTM zone 15N",GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]],PROJECTION["Transverse_Mercator"],PARAMETER["latitude_of_origin",0],PARAMETER["central_meridian",-93],PARAMETER["scale_factor",0.9996],PARAMETER["false_easting",500000],PARAMETER["false_northing",0],UNIT["metre",1],AXIS["Easting",EAST],AXIS["Northing",NORTH],AUTHORITY["EPSG","32615"]]</SRS>
  <GeoTransform> -6.4500000000000000e+01,  1.0000000000000000e+00,  0.0000000000000000e+00,  6.6500000000000000e+01,  0.0000000000000000e+00, -1.0000000000000000e+00</GeoTransform>
  <VRTRasterBand dataType="CFloat32" band="1">
    <SimpleSource>
      <SourceFilename relativeToVRT="0">compressed_20220101_20220101_20220104.tif</SourceFilename>
      <SourceBand>1</SourceBand>
      <SrcRect xOff="0" yOff="0" xSize="128" ySize="128" />
      <DstRect xOff="0" yOff="0" xSize="128" ySize="128" />
    </SimpleSource>
  </VRTRasterBand>
  <VRTRasterBand dataType="CFloat32" band="1">
    <SimpleSource>
      <SourceFilename relativeToVRT="0">NETCDF:"t087_185684_iw2_20220102.h5":"//data/VV"</SourceFilename>
      <SourceBand>1</SourceBand>
      <SrcRect xOff="0" yOff="0" xSize="128" ySize="128" />
      <DstRect xOff="0" yOff="0" xSize="128" ySize="128" />
    </SimpleSource>
  </VRTRasterBand>
  <VRTRasterBand dataType="CFloat32" band="2">
    <SimpleSource>
      <SourceFilename relativeToVRT="0">NETCDF:"t087_185684_iw2_20220103.h5":"//data/VV"</SourceFilename>
      <SourceBand>1</SourceBand>
      <SrcRect xOff="0" yOff="0" xSize="128" ySize="128" />
      <DstRect xOff="0" yOff="0" xSize="128" ySize="128" />
    </SimpleSource>
  </VRTRasterBand>
  <VRTRasterBand dataType="CFloat32" band="3">
    <SimpleSource>
      <SourceFilename relativeToVRT="0">NETCDF:"t087_185684_iw2_20220104.h5":"//data/VV"</SourceFilename>
      <SourceBand>1</SourceBand>
      <SrcRect xOff="0" yOff="0" xSize="128" ySize="128" />
      <DstRect xOff="0" yOff="0" xSize="128" ySize="128" />
    </SimpleSource>
  </VRTRasterBand>
</VRTDataset>
"""  # noqa: E501


def test_parse_vrt(tmp_path, test_vrt):
    with open(tmp_path / "t.vrt", "w") as f:
        f.write(test_vrt)

    filepaths, sds = _parse_vrt_file(tmp_path / "t.vrt")
    assert filepaths == [
        "compressed_20220101_20220101_20220104.tif",
        "t087_185684_iw2_20220102.h5",
        "t087_185684_iw2_20220103.h5",
        "t087_185684_iw2_20220104.h5",
    ]
    assert sds == "data/VV"


class TestEagerLoader:
    def test_iter_blocks(self, tiled_raster_100_by_200):
        # Try the whole raster
        bs = (100, 200)
        loader = EagerLoader(
            reader=RasterReader.from_file(tiled_raster_100_by_200), block_shape=bs
        )
        # `list` should try to load all at once`
        block_slice_tuples = list(loader.iter_blocks())
        assert not loader._thread.is_alive()
        assert len(block_slice_tuples) == 1
        blocks, slices = zip(*list(block_slice_tuples))
        assert blocks[0].shape == (100, 200)
        rows, cols = slices[0]
        assert rows == slice(0, 100)
        assert cols == slice(0, 200)

        # now one block at a time
        bs = (32, 32)
        reader = RasterReader.from_file(tiled_raster_100_by_200)
        loader = EagerLoader(reader=reader, block_shape=bs)
        blocks, slices = zip(*list(loader.iter_blocks()))

        row_blocks = 100 // 32 + 1
        col_blocks = 200 // 32 + 1
        expected_num_blocks = row_blocks * col_blocks
        assert len(blocks) == expected_num_blocks
        assert blocks[0].shape == (32, 32)
        # at the ends, the block_slice_tuples are smaller
        assert blocks[6].shape == (32, 8)
        assert blocks[-1].shape == (4, 8)

    def test_iter_blocks_rowcols(self, tiled_raster_100_by_200):
        # Block size that is a multiple of the raster size
        reader = RasterReader.from_file(tiled_raster_100_by_200)
        loader = EagerLoader(reader=reader, block_shape=(10, 20))
        blocks, slices = zip(*list(loader.iter_blocks()))

        assert blocks[0].shape == (10, 20)
        for rs, cs in slices:
            assert rs.stop - rs.start == 10
            assert cs.stop - cs.start == 20
        loader.notify_finished()

        # Non-multiple block size
        reader = RasterReader.from_file(tiled_raster_100_by_200)
        loader = EagerLoader(reader=reader, block_shape=(32, 32))
        blocks, slices = zip(*list(loader.iter_blocks()))
        assert blocks[0].shape == (32, 32)
        for b, (rs, cs) in zip(blocks, slices):
            assert b.shape == (rs.stop - rs.start, cs.stop - cs.start)
        loader.notify_finished()

    def test_iter_nodata(
        self,
        raster_with_nan,
        raster_with_nan_block,
        raster_with_zero_block,
        tiled_raster_100_by_200,
    ):
        # load one block at a time
        bs = (100, 200)
        reader = RasterReader.from_file(tiled_raster_100_by_200)
        loader = EagerLoader(reader=reader, block_shape=bs)
        blocks, slices = zip(*list(loader.iter_blocks()))
        loader.notify_finished()

        bs = (32, 32)
        row_blocks = 100 // 32 + 1
        col_blocks = 200 // 32 + 1
        expected_num_blocks = row_blocks * col_blocks
        loader = EagerLoader(reader=reader, block_shape=bs)
        blocks, slices = zip(*list(loader.iter_blocks()))
        assert len(blocks) == expected_num_blocks
        assert blocks[0].shape == bs

        # One nan should be fine, will get loaded
        reader = RasterReader.from_file(raster_with_nan)
        loader = EagerLoader(reader=reader, block_shape=bs)
        blocks, slices = zip(*list(loader.iter_blocks()))
        loader.notify_finished()
        assert len(blocks) == expected_num_blocks

        # Now check entire block for a skipped block
        reader = RasterReader.from_file(raster_with_nan_block)
        loader = EagerLoader(reader=reader, block_shape=bs)
        blocks, slices = zip(*list(loader.iter_blocks()))
        loader.notify_finished()
        assert len(blocks) == expected_num_blocks - 1

        # Now check entire block for a skipped block
        reader = RasterReader.from_file(raster_with_zero_block)
        loader = EagerLoader(reader=reader, block_shape=bs)
        blocks, slices = zip(*list(loader.iter_blocks()))
        loader.notify_finished()
        assert len(blocks) == expected_num_blocks - 1



from __future__ import annotations

import logging
from contextlib import ExitStack
from pathlib import Path
from typing import Optional

import numpy as np

from dolphin._types import Filename
from dolphin.io._core import DEFAULT_TIFF_OPTIONS_RIO
from dolphin.utils import full_suffix

from ._constants import CONNCOMP_SUFFIX, DEFAULT_CCL_NODATA, DEFAULT_UNW_NODATA
from ._utils import _zero_from_mask

__all__ = [
    "unwrap_whirlwind",
]


logger = logging.getLogger(__name__)


def unwrap_whirlwind(
    ifg_filename: Filename,
    corr_filename: Filename,
    unw_filename: Filename,
    nlooks: float,
    mask_file: Optional[Filename] = None,
    zero_where_masked: bool = False,
    unw_nodata: Optional[float] = DEFAULT_UNW_NODATA,
    ccl_nodata: Optional[int] = DEFAULT_CCL_NODATA,
    scratchdir: Optional[Filename] = None,
) -> tuple[Path, Path]:
    """Unwrap an interferogram using `whirlwind`.

    Parameters
    ----------
    ifg_filename : Filename
        Path to input interferogram.
    corr_filename : Filename
        Path to input correlation file.
    unw_filename : Filename
        Path to output unwrapped phase file.
    nlooks : float
        Effective number of looks used to form the input correlation data.
    mask_file : Filename, optional
        Path to binary byte mask file, by default None.
        Assumes that 1s are valid pixels and 0s are invalid.
    zero_where_masked : bool, optional
        Set wrapped phase/correlation to 0 where mask is 0 before unwrapping.
        If not mask is provided, this is ignored.
        By default True.
    unw_nodata : float, optional
        If providing `unwrap_callback`, provide the nodata value for your
        unwrapping function.
    ccl_nodata : float, optional
        Nodata value for the connected component labels.
    scratchdir : Filename, optional
        If provided, uses a scratch directory to save the intermediate files
        during unwrapping.

    Returns
    -------
    unw_path : Path
        Path to output unwrapped phase file.
    conncomp_path : Path
        Path to output connected component label file.

    """
    import snaphu
    import whirlwind as ww

    # Create a context manager that combines other context managers -- one for each
    # input raster file. Upon exiting the context block, each context manager in the
    # stack will be closed in LIFO order.
    with ExitStack() as stack:
        if zero_where_masked and (mask_file is not None):
            logger.info(f"Zeroing phase/corr of pixels masked in {mask_file}")
            zeroed_ifg_file, zeroed_corr_file = _zero_from_mask(
                ifg_filename, corr_filename, mask_file
            )
            igram = stack.enter_context(snaphu.io.Raster(zeroed_ifg_file))
            corr = stack.enter_context(snaphu.io.Raster(zeroed_corr_file))
        else:
            igram = stack.enter_context(snaphu.io.Raster(ifg_filename))
            corr = stack.enter_context(snaphu.io.Raster(corr_filename))

        if mask_file is None:
            mask = None
        else:
            mask = stack.enter_context(snaphu.io.Raster(mask_file))

        logger.info("Unwrapping using whirlwind")
        # FIXME: Ad hoc kludge to prevent NaN's in whirlwind cost computation.
        # Remove this when the issue is fixed upstream.
        nlooks = np.clip(nlooks / 2, 1.0, 20.0)
        unw = ww.unwrap(igram, corr, nlooks, mask=mask)

        logger.info("Writing unwrapped phase to raster file")
        with snaphu.io.Raster.create(
            unw_filename,
            like=igram,
            nodata=unw_nodata,
            dtype=np.float32,
            **DEFAULT_TIFF_OPTIONS_RIO,
        ) as unw_raster:
            unw_raster[:, :] = unw

        unw_suffix = full_suffix(unw_filename)
        cc_filename = str(unw_filename).replace(unw_suffix, CONNCOMP_SUFFIX)

        # XXX Whirlwind does not yet provide connected components. Instead, grow
        # connected components using SNAPHU's 'SMOOTH' cost function for now.
        logger.info("Growing connected component labels using SNAPHU")
        with snaphu.io.Raster.create(
            cc_filename,
            like=igram,
            nodata=ccl_nodata,
            dtype=np.uint16,
            **DEFAULT_TIFF_OPTIONS_RIO,
        ) as conncomp:
            snaphu.grow_conncomps(
                unw=unw,
                corr=corr,
                nlooks=nlooks,
                mask=mask,
                cost="smooth",
                scratchdir=scratchdir,
                conncomp=conncomp,
            )

    if zero_where_masked and (mask_file is not None):
        logger.info(f"Zeroing unw/conncomp of pixels masked in {mask_file}")
        return _zero_from_mask(unw_filename, cc_filename, mask_file)

    return Path(unw_filename), Path(cc_filename)



import numpy as np
from numba import njit, stencil
from numpy.typing import ArrayLike
from scipy.ndimage import gaussian_filter

TWOPI = 2 * np.pi


@njit(nogil=True)
def compute_phase_diffs(phase):
    """Compute the total number phase jumps > pi between adjacent pixels.

    If part of `phase` is known to be bad phase (e.g. over water),
    the values should be set to zero or a masked array should be passed:

        unwrapping_error(np.ma.masked_where(bad_area_mask, phase))


    Parameters
    ----------
    phase : ArrayLike
        Unwrapped interferogram phase.

    Returns
    -------
    int
        Total number of jumps exceeding pi.

    """
    return _compute_phase_diffs(phase)


@stencil
def _compute_phase_diffs(phase):
    d1 = np.abs(phase[0, 0] - phase[0, 1]) / np.pi
    d2 = np.abs(phase[0, 0] - phase[1, 0]) / np.pi
    # Subtract 0.5 so that anything below 1 gets rounded to 0
    return round(d1 - 0.5) + round(d2 - 0.5)


def rewrap_to_twopi(arr: ArrayLike) -> np.ndarray:
    """Rewrap `arr` to be between [-pi and pi].

    Parameters
    ----------
    arr : ArrayLike
        Unwrapped floating point phase

    Returns
    -------
    np.ndarray
        Array of phases between -pi and pi

    """
    return np.mod(np.pi + arr, TWOPI) - np.pi


def gaussian_filter_nan(
    image: ArrayLike, sigma: float, mode="constant", **kwargs
) -> np.ndarray:
    """Apply a gaussian filter to an image with NaNs (avoiding all nans).

    The scipy.ndimage `gaussian_filter` will make the output all NaNs if
    any of the pixels in the input that touches the kernel is NaN

    Source:
    https://stackoverflow.com/a/36307291

    Parameters
    ----------
    image : ndarray
        Image with nans to filter
    sigma : float
        Size of filter kernel. passed into `gaussian_filter`
    mode : str, default = "constant"
        Boundary mode for `[scipy.ndimage.gaussian_filter][]`
    **kwargs : Any
        Passed into `[scipy.ndimage.gaussian_filter][]`

    Returns
    -------
    ndarray
        Filtered version of `image`.

    """
    if np.sum(np.isnan(image)) == 0:
        return gaussian_filter(image, sigma=sigma, mode=mode, **kwargs)

    V = image.copy()
    nan_idxs = np.isnan(image)
    V[nan_idxs] = 0
    V_filt = gaussian_filter(V, sigma, **kwargs)

    W = np.ones(image.shape)
    W[nan_idxs] = 0
    W_filt = gaussian_filter(W, sigma, **kwargs)

    return V_filt / W_filt


def _get_ambiguities(unw: ArrayLike, round_decimals: int = 4) -> np.ndarray:
    mod_2pi_image = np.mod(np.pi + unw, TWOPI) - np.pi
    re_wrapped = np.round(mod_2pi_image, round_decimals)
    return np.round((unw - re_wrapped) / (TWOPI), round_decimals - 1)


def _fill_masked_ambiguities(
    amb_image: ArrayLike, mask: ArrayLike, filter_sigma: int = 60
) -> np.ndarray:
    masked_ambs = amb_image.copy()
    masked_ambs[mask] = np.nan
    ambs_filled = np.round(gaussian_filter_nan(amb_image, filter_sigma))

    out_filled = amb_image.copy()
    out_filled[mask] = ambs_filled[mask]
    return out_filled


def _smooth_masked_areas(unw, mask, filter_sigma: int = 60):
    amb = _get_ambiguities(unw)
    amb_filled = _fill_masked_ambiguities(amb, mask, filter_sigma=filter_sigma)

    out = unw.copy()
    rewrapped_phase_vec = np.mod(np.pi + unw[mask], TWOPI) - np.pi
    new_amb_vec = amb_filled[mask]
    out[mask] = rewrapped_phase_vec + (new_amb_vec * TWOPI)
    return out



from __future__ import annotations

import logging
from pathlib import Path
from typing import Annotated, Any, Optional

from opera_utils import get_dates, sort_files_by_date
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    StringConstraints,
    field_validator,
    model_validator,
)

from dolphin._types import TropoModel, TropoType

from ._common import (
    InputOptions,
    InterferogramNetwork,
    OutputOptions,
    PhaseLinkingOptions,
    PsOptions,
    TimeseriesOptions,
    WorkflowBase,
    _read_file_list_or_glob,
)
from ._unwrap_options import UnwrapOptions

__all__ = [
    "DisplacementWorkflow",
    "CorrectionOptions",
]

logger = logging.getLogger(__name__)


# Add a class for troposphere, ionosphere corrections, with geometry files and DEM
class CorrectionOptions(BaseModel, extra="forbid"):
    """Configuration for the auxillary phase corrections."""

    _atm_directory: Path = Path("atmosphere")
    _iono_date_fmt: list[str] = ["%j0.%y", "%Y%j0000"]

    troposphere_files: list[Path] = Field(
        default_factory=list,
        description=(
            "List of weather-model files (one per date) for tropospheric corrections"
        ),
    )

    tropo_date_fmt: str = Field(
        "%Y%m%d",
        description="Format of dates contained in weather-model filenames",
    )

    tropo_package: Annotated[str, StringConstraints(to_lower=True)] = Field(
        "pyaps",
        description="Package for tropospheric correction. Choices: pyaps, raider",
    )

    tropo_model: TropoModel = Field(
        TropoModel.ECMWF, description="source of the atmospheric model."
    )

    tropo_delay_type: TropoType = Field(
        TropoType.COMB,
        description=(
            "Tropospheric delay type to calculate, comb contains both wet "
            "and dry delays."
        ),
    )

    ionosphere_files: list[Path] = Field(
        default_factory=list,
        description=(
            "List of GNSS-derived TEC maps for ionospheric corrections (one per date)."
            " Source is https://cddis.nasa.gov/archive/gnss/products/ionex/"
        ),
    )

    geometry_files: list[Path] = Field(
        default_factory=list,
        description=(
            "Line-of-sight geometry files for each burst/SLC stack area, for use in"
            " correction computations."
        ),
    )
    dem_file: Optional[Path] = Field(
        None,
        description="DEM file for tropospheric/ topographic phase corrections.",
    )

    @field_validator(
        "troposphere_files", "ionosphere_files", "geometry_files", mode="before"
    )
    @classmethod
    def _to_empty_list(cls, v):
        return v if v is not None else []


class DisplacementWorkflow(WorkflowBase):
    """Configuration for the workflow."""

    # Paths to input/output files
    input_options: InputOptions = Field(default_factory=InputOptions)
    cslc_file_list: list[Path] = Field(
        default_factory=list,
        description=(
            "list of CSLC files, or newline-delimited file "
            "containing list of CSLC files."
        ),
    )
    output_options: OutputOptions = Field(default_factory=OutputOptions)

    # Options for each step in the workflow
    ps_options: PsOptions = Field(default_factory=PsOptions)
    amplitude_dispersion_files: list[Path] = Field(
        default_factory=list,
        description=(
            "Paths to existing Amplitude Dispersion file (1 per SLC region) for PS"
            " update calculation. If none provided, computed using the input SLC stack."
        ),
    )
    amplitude_mean_files: list[Path] = Field(
        default_factory=list,
        description=(
            "Paths to an existing Amplitude Mean files (1 per SLC region) for PS update"
            " calculation. If none provided, computed using the input SLC stack."
        ),
    )

    phase_linking: PhaseLinkingOptions = Field(default_factory=PhaseLinkingOptions)
    interferogram_network: InterferogramNetwork = Field(
        default_factory=InterferogramNetwork
    )
    unwrap_options: UnwrapOptions = Field(default_factory=UnwrapOptions)
    timeseries_options: TimeseriesOptions = Field(default_factory=TimeseriesOptions)
    correction_options: CorrectionOptions = Field(default_factory=CorrectionOptions)

    # internal helpers
    # Stores the list of directories to be created by the workflow
    model_config = ConfigDict(
        extra="allow", json_schema_extra={"required": ["cslc_file_list"]}
    )

    # validators
    # reuse the _read_file_list_or_glob
    _check_cslc_file_glob = field_validator("cslc_file_list", mode="before")(
        _read_file_list_or_glob
    )

    @model_validator(mode="after")
    def _check_input_files_exist(self) -> DisplacementWorkflow:
        file_list = self.cslc_file_list
        if not file_list:
            msg = "Must specify list of input SLC files."
            raise ValueError(msg)

        input_options = self.input_options
        date_fmt = input_options.cslc_date_fmt
        # Filter out files that don't have dates in the filename
        files_matching_date = [Path(f) for f in file_list if get_dates(f, fmt=date_fmt)]
        if len(files_matching_date) < len(file_list):
            msg = (
                f"Found {len(files_matching_date)} files with dates like {date_fmt} in"
                f" the filename out of {len(file_list)} files."
            )
            raise ValueError(msg)

        ext = file_list[0].suffix
        # If they're HDF5/NetCDF files, we need to check that the subdataset exists
        if ext in [".h5", ".nc"]:
            subdataset = input_options.subdataset
            if subdataset is None:
                msg = "Must provide subdataset name for input NetCDF/HDF5 files."
                raise ValueError(msg)

        # Coerce the file_list to a sorted list of Path objects
        self.cslc_file_list = [
            Path(f) for f in sort_files_by_date(file_list, file_date_fmt=date_fmt)[0]
        ]

        return self

    def model_post_init(self, __context: Any) -> None:
        """After validation, set up properties for use during workflow run."""
        super().model_post_init(__context)
        # Ensure outputs from workflow steps are within work directory.
        if not self.keep_paths_relative:
            # Resolve all CSLC paths:
            self.cslc_file_list = [p.resolve(strict=False) for p in self.cslc_file_list]

        work_dir = self.work_directory
        # For each workflow step that has an output folder, move it inside
        # the work directory (if it's not already inside).
        # They may already be inside if we're loading from a json/yaml file.
        for step in [
            "ps_options",
            "phase_linking",
            "interferogram_network",
            "unwrap_options",
            "timeseries_options",
        ]:
            opts = getattr(self, step)
            if isinstance(opts, dict):
                # If this occurs, we are printing the schema.
                # Using newer pydantic `model_construct`, this would be a dict,
                # instead of an object.
                # We don't care about the subsequent logic here
                return

            if opts._directory.parent != work_dir:
                opts._directory = work_dir / opts._directory
            if not self.keep_paths_relative:
                opts._directory = opts._directory.resolve(strict=False)

        # Track the directories that need to be created at start of workflow
        self._directory_list = [
            work_dir,
            self.ps_options._directory,
            self.phase_linking._directory,
            self.interferogram_network._directory,
            self.unwrap_options._directory,
            self.timeseries_options._directory,
        ]
        # Add the output PS files we'll create to the `PS` directory, making
        # sure they're inside the work directory
        ps_opts = self.ps_options
        ps_opts._amp_dispersion_file = work_dir / ps_opts._amp_dispersion_file
        ps_opts._amp_mean_file = work_dir / ps_opts._amp_mean_file
        ps_opts._output_file = work_dir / ps_opts._output_file

        self.timeseries_options._velocity_file = (
            work_dir / self.timeseries_options._velocity_file
        )



from pathlib import Path

import numpy as np
import pytest

from dolphin import io
from dolphin.unwrap import _utils

# Dataset has no geotransform, gcps, or rpcs. The identity matrix will be returned.
pytestmark = pytest.mark.filterwarnings(
    "ignore::rasterio.errors.NotGeoreferencedWarning",
    "ignore:.*io.FileIO.*:pytest.PytestUnraisableExceptionWarning",
)


@pytest.fixture()
def corr_raster(raster_100_by_200):
    # Make a correlation raster of all 1s in the same directory as the raster
    d = Path(raster_100_by_200).parent
    arr = np.ones((100, 200), dtype="float32")
    # The first 20 rows have nodata
    arr[:20, :] = np.nan
    filename = d / "corr_raster.cor.tif"
    io.write_arr(
        arr=arr,
        output_name=filename,
        like_filename=raster_100_by_200,
        nodata=np.nan,
        driver="GTiff",
    )
    return filename


@pytest.fixture()
def mask_raster(raster_100_by_200):
    # Make a correlation raster of all 1s in the same directory as the raster
    d = Path(raster_100_by_200).parent
    filename = d / "mask.tif"
    arr = np.ones((100, 200), dtype=np.uint8)
    # Mask the first 20 columns
    arr[:, :20] = 0
    io.write_arr(
        arr=arr,
        output_name=filename,
        like_filename=raster_100_by_200,
        driver="GTiff",
    )
    return filename


def test_create_combined_mask(corr_raster, mask_raster):
    out_raster = _utils.create_combined_mask(
        mask_filename=mask_raster, image_filename=corr_raster
    )
    assert Path(out_raster).name == "combined_mask.tif"
    mask = io.load_gdal(out_raster)
    assert (mask[:20] == 0).all()
    assert (mask[:, :20] == 0).all()
    assert (mask[20:, 20:] == 1).all()


def test_set_nodata_values(corr_raster):
    import rasterio

    corr_copy = corr_raster.parent / "corr_copy.tif"
    # The first 20 cols have nodata
    arr = np.ones((100, 200), dtype="float32")
    arr[:, :20] = -1
    io.write_arr(
        arr=arr,
        output_name=corr_copy,
        like_filename=corr_raster,
        nodata=-1,
        driver="GTiff",
    )
    with rasterio.open(corr_copy) as src:
        assert src.nodata == -1
        a = src.read(1)
        assert np.all(a[:, :20] == -1)
        assert np.all(a[:, 20:] == 1)

    _utils.set_nodata_values(filename=corr_copy, like_filename=corr_raster)
    # Without specifying `output_nodata`, should keep existing nodata of `filename`
    with rasterio.open(corr_copy) as src:
        assert src.nodata == -1
        a = src.read(1)
        assert np.all(a[:20] == -1)
        assert np.all(a[:, :20] == -1)

    _utils.set_nodata_values(
        filename=corr_copy, output_nodata=-2, like_filename=corr_raster
    )
    with rasterio.open(corr_copy) as src:
        assert src.nodata == -2
        a = src.read(1)
        assert np.all(a[:20] == -2)
        assert np.all(a[:, :20] == -2)



import contextlib
from pathlib import Path

import pytest

from dolphin.cli import main


@pytest.mark.parametrize("option", ["-h", "--help"])
def test_help(capsys, option):
    with contextlib.suppress(SystemExit):
        main([option])
    output = capsys.readouterr().out
    assert " dolphin [-h] [--version] {run,config,unwrap,timeseries,filter}" in output


def test_empty(capsys):
    with contextlib.suppress(SystemExit):
        main([])
    output = capsys.readouterr().out
    assert " dolphin [-h] [--version] {run,config,unwrap,timeseries,filter}" in output


@pytest.mark.parametrize("sub_cmd", ["run", "config", "filter", "unwrap", "timeseries"])
@pytest.mark.parametrize("option", ["-h", "--help"])
def test_subcommand_help(capsys, sub_cmd, option):
    with contextlib.suppress(SystemExit):
        main([sub_cmd, option])
    output = capsys.readouterr().out
    assert f"usage: dolphin {sub_cmd} [-h]" in output


def test_cli_config_basic(tmpdir, slc_file_list):
    with tmpdir.as_cwd(), contextlib.suppress(SystemExit):
        main(
            [
                "config",
                "--threads-per-worker",
                "2",
                "--strides",
                "6",
                "3",
                "--slc-files",
                *list(map(str, slc_file_list)),
            ]
        )
        assert Path("dolphin_config.yaml").exists()



from __future__ import annotations

import functools
import logging
import shutil
import tempfile
from pathlib import Path
from typing import Any, Callable

logger = logging.getLogger(__name__)

__all__ = [
    "atomic_output",
]


def atomic_output(
    output_arg: str = "output_file",
    is_dir: bool = False,
    use_tmp: bool = False,
    overwrite: bool = False,
) -> Callable:
    """Use a temporary file/directory for the `output_arg` until the function finishes.

    Decorator is used on a function which writes to an output file/directory in blocks.
    If the function were interrupted, the file/directory would be partially complete.

    This decorator replaces the final output name with a temp file/dir, and then
    renames the temp file/dir to the final name after the function finishes.

    Note that when `is_dir=True`, `output_arg` can be a directory (if multiple files
    are being written to). In this case, the entire directory is temporary, and
    renamed after the function finishes.

    Parameters
    ----------
    output_arg : str, optional
        The name of the argument to replace, by default 'output_file'
    is_dir : bool, default = False
        If `True`, the output argument is a directory, not a file
    use_tmp : bool, default = False
        If `False`, uses the parent directory of the desired output, with
        a random suffix added to the name to distinguish from actual output.
        If `True`, uses the `/tmp` directory (or wherever the default is
        for the `tempfile` module).
    overwrite : bool, default = False
        Overwrite an existing file.
        If `False` raises `FileExistsError` if the file already exists.

    Returns
    -------
    Callable
        The decorated function

    Raises
    ------
    FileExistsError
        if the file for `output_arg` already exists (if out_dir=False), or
        if the directory at `output_arg` exists and is non-empty.

    Notes
    -----
    The output at `output_arg` *must not* exist already, or the decorator will error
    (though if `is_dir=True`, it is allowed to be an empty directory).
    The function being decorated *must* be called with keyword args for `output_arg`.

    """

    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            # Extract the output file path
            if kwargs.get(output_arg):
                final_out_name = kwargs[output_arg]
            else:
                msg = (
                    f"Argument {output_arg} not passed to function {func.__name__}:"
                    f" {kwargs}"
                )
                raise FileExistsError(msg)

            final_path = Path(final_out_name)
            # Make sure the desired final output doesn't already exist
            _raise_if_exists(final_path, is_dir=is_dir, overwrite=overwrite)
            # None means that tempfile will use /tmp
            tmp_dir = final_path.parent if not use_tmp else None

            # Make the tempfile start the same as the desired output
            prefix = final_path.name
            if is_dir:
                # Create a temporary directory
                temp_path = tempfile.mkdtemp(dir=tmp_dir, prefix=prefix)
            else:
                # Create a temporary file
                suffix = final_path.suffix
                _, temp_path = tempfile.mkstemp(
                    dir=tmp_dir, prefix=prefix, suffix=suffix
                )
            logger.debug("Writing to temp file %s instead of %s", temp_path, final_path)

            try:
                # Replace the output file path with the temp file
                # It would be like this if we only allows keyword:
                kwargs[output_arg] = temp_path
                # Execute the original function
                result = func(*args, **kwargs)
                # Move the temp file to the final location
                logger.debug("Moving %s to %s", temp_path, final_path)
                shutil.move(temp_path, final_path)

                return result
            finally:
                logger.debug("Cleaning up temp file %s", temp_path)
                # Different cleanup is needed
                if is_dir:
                    shutil.rmtree(temp_path, ignore_errors=True)
                else:
                    Path(temp_path).unlink(missing_ok=True)

        return wrapper

    return decorator


def _raise_if_exists(final_path: Path, is_dir: bool, overwrite: bool):
    msg = f"{final_path} already exists"
    if final_path.exists():
        logger.debug(f"{final_path} already exists")
        if overwrite:
            if final_path.is_dir():
                shutil.rmtree(final_path)
            else:
                final_path.unlink()
            return

        if is_dir and final_path.is_dir():
            # We can work with an empty directory
            try:
                final_path.rmdir()
            except OSError as e:
                err_msg = str(e)
                if "Directory not empty" in err_msg:
                    raise FileExistsError(msg) from e
                else:
                    # Some other error we don't know
                    raise
        else:
            raise FileExistsError(msg)



"""Compute Generalized Likelihood Ratio Test (GLRT) cutoffs.

Runs over different sample sizes (N), significance levels (alpha), and scales.
Results are saved in a CSV file.
"""

from __future__ import annotations

from itertools import product

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from numpy import log as ln
from scipy.stats import rayleigh


def get_test_stat_glrt(
    N: int,
    scale: float = 10,
    nsim: int = 500,
) -> np.ndarray:
    """Compute test statistics for the GLRT.

    Parameters
    ----------
    N : int
        Number of observations
    scale : float
        Scale parameter for the Rayleigh distribution
    nsim : float
        Number of simulations

    Returns
    -------
    ndarray:
        A numpy array of test statistics

    """
    x = rayleigh.rvs(scale=scale, size=(nsim, 2 * N))

    scale2_p = (x[:, :N] ** 2).mean(axis=1) / 2
    scale2_q = (x[:, N:] ** 2).mean(axis=1) / 2
    scale2_pooled = (scale2_p + scale2_q) / 2
    return 2 * ln(scale2_pooled) - ln(scale2_p) - ln(scale2_q)


def get_alpha_cutoff(
    alpha: float, N: int, nsim: int = 50000, scale: float = 10
) -> float:
    """Compute alpha cutoff for the GLRT test.

    Parameters
    ----------
    alpha : float
        Significance level
    N : int
        Number of observations
    scale : float
        Scale parameter for the Rayleigh distribution
    nsim : float
        Number of simulations


    Returns
    -------
    float:
        Alpha cutoff value

    """
    return np.percentile(
        get_test_stat_glrt(N=N, nsim=nsim, scale=scale), 100 * (1 - alpha)
    )


def compute_cutoff(
    alpha: float, N: int, scale: float
) -> tuple[int, float, float, float]:
    """Run computation for a single combination of N, alpha, and scale.

    Parameters
    ----------
    alpha : float
        Significance level
    N : int
        Number of observations
    scale : float
        Scale parameter for the Rayleigh distribution

    Returns
    -------
    (N, alpha, scale, computed cutoff)

    """
    return (N, alpha, scale, get_alpha_cutoff(alpha=alpha, N=N, scale=scale))


if __name__ == "__main__":
    Narr = list(range(1, 301))
    scales = (1, 10, 50)
    alphas = [0.05, 0.01, 0.005, 0.001]

    results = Parallel(n_jobs=30)(
        delayed(compute_cutoff)(*row) for row in product(Narr, alphas, scales)
    )

    df = pd.DataFrame(data=results, columns=["N", "alpha", "scale", "cutoff"])
    d2 = df.groupby(["N", "alpha"]).mean().round(4)[["cutoff"]]
    d2.to_csv("dolphin/shp/glrt_cutoffs.csv", index=True)



import numpy as np
import pytest

from dolphin._types import HalfWindow, Strides
from dolphin.io._blocks import (
    BlockIndices,
    StridedBlockManager,
    get_slice_length,
    iter_blocks,
)
from dolphin.utils import compute_out_shape, upsample_nearest


def test_block_indices_create():
    b = BlockIndices(0, 3, 1, 5)
    assert b.row_start == 0
    assert b.row_stop == 3
    assert b.col_start == 1
    assert b.col_stop == 5
    assert tuple(b) == (slice(0, 3, None), slice(1, 5, None))


def test_compute_out_size():
    strides = Strides(1, 1)
    assert compute_out_shape((6, 6), strides) == (6, 6)

    strides = Strides(3, 3)
    assert compute_out_shape((6, 6), strides) == (2, 2)

    # 1,2 more in each direction shouldn't change it
    assert compute_out_shape((7, 7), strides) == (2, 2)
    assert compute_out_shape((8, 8), strides) == (2, 2)

    # 1,2 fewer should bump down to 1
    assert compute_out_shape((5, 5), strides) == (1, 1)
    assert compute_out_shape((4, 4), strides) == (1, 1)


def test_iter_blocks():
    out_blocks = iter_blocks((3, 5), (2, 2))
    assert hasattr(out_blocks, "__iter__")
    assert list(out_blocks) == [
        BlockIndices(0, 2, 0, 2),
        BlockIndices(0, 2, 2, 4),
        BlockIndices(0, 2, 4, 5),
        BlockIndices(2, 3, 0, 2),
        BlockIndices(2, 3, 2, 4),
        BlockIndices(2, 3, 4, 5),
    ]


@pytest.mark.parametrize("block_shape", [(5, 5), (10, 20), (13, 27)])
def test_iter_blocks_coverage(block_shape):
    shape = (100, 200)
    check_out = np.zeros(shape)

    for rs, cs in iter_blocks(shape, block_shape):
        check_out[rs, cs] += 1

    # Everywhere should have been touched once by the iteration
    assert np.all(check_out == 1)


def test_iter_blocks_overlap():
    # Block size that is a multiple of the raster size
    shape = (100, 200)
    check_out = np.zeros(shape)

    for rs, cs in iter_blocks(shape, (30, 30), overlaps=(5, 5)):
        check_out[rs, cs] += 1

    # Everywhere should have been touched *at least* once by the iteration
    assert np.all(check_out >= 1)


def test_iter_blocks_offset_margin():
    # Block size that is a multiple of the raster size
    shape = (100, 200)
    check_out = np.zeros(shape)

    for rs, cs in iter_blocks(shape, (30, 30), start_offsets=(2, 3)):
        check_out[rs, cs] += 1

    # Everywhere should have been touched once by the iteration
    assert np.all(check_out[2:, 3:] == 1)
    # offset should still be 0
    assert np.all(check_out[:2, :3] == 0)

    check_out[:] = 0
    for rs, cs in iter_blocks(shape, (30, 30), end_margin=(4, 5)):
        check_out[rs, cs] += 1
    # Everywhere except the end should be 1
    assert np.all(check_out[:4, :5] == 1)
    assert np.all(check_out[-4:, -5:] == 0)


def test_nonzero_block_size_with_margin():
    shape = (33, 67)
    block_shape = (5, 5)
    offset = margin = (0, 1)
    check_out = np.zeros(shape)
    for rs, cs in iter_blocks(
        shape, block_shape, start_offsets=offset, end_margin=margin
    ):
        assert get_slice_length(rs) > 0
        assert get_slice_length(cs) > 0
        check_out[rs, cs] += 1
    assert np.all(check_out[:, 1:-1] == 1)


class TestBlockManager:
    def test_basic(self):
        # Check no stride version
        bm = StridedBlockManager((5, 5), (2, 3))
        assert list(bm.iter_outputs()) == [
            BlockIndices(row_start=0, row_stop=2, col_start=0, col_stop=3),
            BlockIndices(row_start=0, row_stop=2, col_start=3, col_stop=5),
            BlockIndices(row_start=2, row_stop=4, col_start=0, col_stop=3),
            BlockIndices(row_start=2, row_stop=4, col_start=3, col_stop=5),
            BlockIndices(row_start=4, row_stop=5, col_start=0, col_stop=3),
            BlockIndices(row_start=4, row_stop=5, col_start=3, col_stop=5),
        ]

        outs, out_trim, ins, in_no_pads, in_trim = zip(*list(bm.iter_blocks()))
        assert outs == ins
        assert outs == in_no_pads
        assert all(
            (rs, cs) == (slice(0, None), slice(0, None)) for (rs, cs) in out_trim
        )

    def test_iter_outputs(self):
        nrows, ncols = (100, 200)
        xs, ys = 3, 3  # strides
        hx, hy = 3, 1  # half window
        bm = StridedBlockManager(
            arr_shape=(nrows, ncols),
            block_shape=(17, 27),
            strides=Strides(ys, xs),
            half_window=HalfWindow(hy, hx),
        )

        out_row_margin = hy // ys
        out_col_margin = hx // ys
        for row_slice, col_slice in bm.iter_outputs():
            assert row_slice.start >= out_row_margin
            assert col_slice.start >= out_col_margin
            assert row_slice.stop < nrows - out_row_margin
            assert col_slice.stop < ncols - out_col_margin


# @pytest.mark.skip(reason="Uses old logic ")
class TestFakeProcess:
    def _fake_process(self, in_arr, strides: Strides, half_window: HalfWindow):
        """Dummy processing which has same nodata pattern as `phase_link`."""
        nrows, ncols = in_arr.shape
        row_half, col_half = half_window.y, half_window.x
        rs, cs = strides.y, strides.x
        out_nrows, out_ncols = compute_out_shape(in_arr.shape, strides=Strides(rs, cs))
        out = np.ones((out_nrows, out_ncols))
        for out_r in range(out_nrows):
            for out_c in range(out_ncols):
                # the input indexes computed from the output idx and strides
                # Note: weirdly, moving these out of the loop causes r_start
                # to be 0 in some cases...
                in_r_start = rs // 2
                in_c_start = cs // 2
                in_r = in_r_start + out_r * rs
                in_c = in_c_start + out_c * cs

                # Check if the window is completely in bounds
                if in_r + row_half >= nrows or in_r - row_half < 0:
                    out[out_r, out_c] = np.nan
                if in_c + col_half >= ncols or in_c - col_half < 0:
                    out[out_r, out_c] = np.nan
        return out

    def fake_process_blocks(
        self, in_shape, half_window: HalfWindow, strides: Strides, block_shape
    ):
        out_shape = compute_out_shape(in_shape, strides)

        # full_res_data = np.random.randn(*in_shape) + 1j * np.random.randn(*in_shape)
        # full_res_data = full_res_data.astype(np.complex64)
        rng = np.random.default_rng()
        full_res_data = rng.normal(size=in_shape).astype("float32")
        out_arr = np.zeros(out_shape, dtype=full_res_data.dtype)
        out_full_res = np.zeros_like(full_res_data)
        counts = np.zeros(out_shape, dtype=int)

        bm = StridedBlockManager(
            in_shape, block_shape=block_shape, strides=strides, half_window=half_window
        )
        for (
            (out_rows, out_cols),
            (out_trim_rows, out_trim_cols),
            (in_rows, in_cols),
            (in_no_pad_rows, in_no_pad_cols),
            (_in_trim_rows, _in_trim_cols),
        ) in bm.iter_blocks():
            in_data = full_res_data[in_rows, in_cols]
            out_data = self._fake_process(in_data, strides, half_window)

            # inner = _get_trimmed_full_res(out_arr, in_block, in_no_pad_block)
            data_trimmed = out_data[out_trim_rows, out_trim_cols]
            assert np.all(~np.isnan(data_trimmed))
            assert get_slice_length(out_rows) == data_trimmed.shape[0]
            assert get_slice_length(out_cols) == data_trimmed.shape[1]

            out_arr[out_rows, out_cols] = data_trimmed
            counts[out_rows, out_cols] += 1

            out_full_nrows = get_slice_length(in_no_pad_rows)
            out_full_ncols = get_slice_length(in_no_pad_cols)
            out_upsampled = upsample_nearest(
                data_trimmed, (out_full_nrows, out_full_ncols)
            )
            out_full_res[in_no_pad_rows, in_no_pad_cols] = out_upsampled

        # Now check the inner part, away from the expected border of zeros
        out_row_margin, out_col_margin = bm.output_margin
        inner = (
            slice(out_row_margin, -out_row_margin),
            slice(out_col_margin, -out_col_margin),
        )
        assert not np.any(out_arr[inner] == 0)
        assert np.all(counts[inner] == 1)

    @pytest.mark.parametrize("in_shape", [(100, 200), (101, 201)])
    @pytest.mark.parametrize(
        "half_window",
        [HalfWindow(x=1, y=1), HalfWindow(x=3, y=1), HalfWindow(x=3, y=3)],
    )
    @pytest.mark.parametrize("strides", [Strides(1, 1), Strides(1, 3), Strides(3, 3)])
    @pytest.mark.parametrize("block_shape", [(15, 15), (20, 30), (17, 27)])
    def test_block_manager_fake_process(
        self, in_shape, half_window, strides, block_shape
    ):
        self.fake_process_blocks(in_shape, half_window, strides, block_shape)

    def test_failing_block_params(self):
        # Extra test from real-data params
        half_window, strides = HalfWindow(x=11, y=5), Strides(6, 3)
        in_shape, block_shape = (2050, 4050), (1024, 1024)
        self.fake_process_blocks(in_shape, half_window, strides, block_shape)



import itertools
from datetime import datetime, timedelta

import numpy as np
import numpy.testing as npt
import pytest
from numpy.linalg import lstsq as lstsq_numpy

from dolphin import io, timeseries
from dolphin.utils import format_dates

NUM_DATES = 10
DT = 12
START_DATE = datetime(2020, 1, 1)
SHAPE = 50, 50
VELO_RAD_PER_YEAR = 5.0
VELO_RAD_PER_DAY = VELO_RAD_PER_YEAR / 365.25


def make_sar_dates():
    sar_date_list = []
    for idx in range(NUM_DATES):
        sar_date_list.append(START_DATE + timedelta(days=idx * DT))
    return sar_date_list


def make_sar_phases(sar_dates):
    out = np.empty((NUM_DATES, *SHAPE), dtype="float32")
    # Make a linear ramp which increases each time interval
    ramp0 = np.ones((SHAPE[0], 1)) * np.arange(SHAPE[1]).reshape(1, -1)
    ramp0 /= SHAPE[1]
    for idx in range(len(sar_dates)):
        cur_slope = VELO_RAD_PER_DAY * DT * idx
        out[idx] = cur_slope * ramp0
    return out


def make_ifg_date_pairs(sar_dates):
    """Form all possible unwrapped interferogram pairs from the date list"""
    return [tuple(pair) for pair in itertools.combinations(sar_dates, 2)]


def make_ifgs(sar_phases):
    """Form all possible unwrapped interferogram pairs from the sar images"""
    return np.stack(
        [(sec - ref) for (ref, sec) in itertools.combinations(sar_phases, 2)]
    )


@pytest.fixture(scope="module")
def data():
    sar_dates = make_sar_dates()
    sar_phases = make_sar_phases(sar_dates)
    ifg_date_pairs = make_ifg_date_pairs(sar_dates)
    ifgs = make_ifgs(sar_phases)
    return sar_dates, sar_phases, ifg_date_pairs, ifgs


def solve_with_removal(A, b):
    good_idxs = (~np.isnan(b)).flatten()
    b_clean = b[good_idxs]
    A_clean = A[good_idxs, :]
    return lstsq_numpy(A_clean, b_clean)[0]


def solve_by_zeroing(A, b):
    weight = (~np.isnan(b)).astype(float)
    A0 = A.copy()
    A0 = A * weight[:, None]
    b2 = np.nan_to_num(b)
    return lstsq_numpy(A0, b2)[0]


class TestUtils:
    def test_datetime_to_float(self):
        sar_dates = make_sar_dates()
        date_arr = timeseries.datetime_to_float(sar_dates)
        expected = np.array(
            [0.0, 12.0, 24.0, 36.0, 48.0, 60.0, 72.0, 84.0, 96.0, 108.0]
        )
        assert np.allclose(date_arr, expected)

    def test_incidence_matrix(self):
        A = timeseries.get_incidence_matrix([(0, 1), (1, 2), (2, 3), (3, 4), (4, 5)])
        assert A.shape == (5, 5)
        expected = np.array(
            [
                [1, 0, 0, 0, 0],
                [-1, 1, 0, 0, 0],
                [0, -1, 1, 0, 0],
                [0, 0, -1, 1, 0],
                [0, 0, 0, -1, 1],
            ]
        )
        np.testing.assert_array_equal(A, expected)


class TestInvert:
    @pytest.fixture
    def A(self, data):
        _, _, ifg_date_pairs, _ = data
        return timeseries.get_incidence_matrix(ifg_date_pairs)

    def test_basic(self, data, A):
        sar_dates, sar_phases, ifg_date_pairs, ifgs = data

        # Get one pixel
        b = ifgs[:, -1, -1]
        weights = np.ones_like(b)
        x, residual = timeseries.weighted_lstsq_single(A, b, weights)
        assert x.shape[0] == len(sar_dates) - 1
        npt.assert_allclose(x, sar_phases[1:, -1, -1], atol=1e-5)
        assert np.array(residual).item() < 1e-5

    def test_stack(self, data, A):
        sar_dates, sar_phases, ifg_date_pairs, ifgs = data

        phi_stack, residuals = timeseries.invert_stack(A, ifgs)
        assert phi_stack.shape == sar_phases[1:].shape
        assert residuals.shape == sar_phases[0].shape
        npt.assert_allclose(phi_stack, sar_phases[1:], atol=1e-5)
        npt.assert_array_less(residuals, 1e-5)

    def test_weighted_stack(self, data, A):
        sar_dates, sar_phases, ifg_date_pairs, ifgs = data

        weights = np.ones_like(ifgs)
        phi, residuals = timeseries.invert_stack(A, ifgs, weights)
        assert phi.shape[0] == len(sar_dates) - 1
        npt.assert_allclose(phi, sar_phases[1:], atol=1e-5)
        # Here there is no noise, so it's over determined
        weights[1, :, :] = 0
        phi2, residuals2 = timeseries.invert_stack(A, ifgs, weights)
        npt.assert_allclose(phi2, sar_phases[1:], atol=1e-5)
        npt.assert_allclose(residuals2, residuals, atol=1e-5)

    def test_remove_row_vs_weighted(self, data, A):
        """Check that removing a row/data point is equivalent to zero-weighting it."""
        sar_dates, sar_phases, ifg_date_pairs, ifgs = data
        dphi = ifgs[:, -1, -1]
        # add noise to the ifgs, or we can't tell a difference
        dphi_noisy = dphi + np.random.normal(0, 0.3, dphi.shape)

        # Once with a zeroed weight
        weights = np.ones_like(dphi_noisy)
        weights[1] = 0
        phi_0, residual_0 = timeseries.weighted_lstsq_single(A, dphi_noisy, weights)

        A2 = A.copy()
        # delete a row
        A2 = np.delete(A2, 1, axis=0)
        # delete the data point
        dphi2 = np.delete(dphi_noisy, 1)
        weights2 = np.delete(weights, 1)
        phi2, residual2 = timeseries.weighted_lstsq_single(A2, dphi2, weights2)
        npt.assert_allclose(phi_0, phi2, atol=1e-5)

        npt.assert_allclose(residual_0, residual2, atol=1e-5)

    @pytest.fixture
    def unw_files(self, tmp_path, data, raster_100_by_200):
        """Write the data to disk and return the file names."""
        sar_dates, sar_phases, ifg_date_pairs, ifgs = data

        out = []
        # use the raster as a template
        for pair, ifg in zip(ifg_date_pairs, ifgs):
            fname = tmp_path / (format_dates(*pair) + ".tif")
            io.write_arr(arr=ifg, output_name=fname, like_filename=raster_100_by_200)
            out.append(fname)

        return out

    def test_invert_unw_network(self, data, unw_files, tmp_path):
        output_dir = tmp_path / "output"
        output_dir.mkdir()
        ref_point = (0, 0)
        out_files = timeseries.invert_unw_network(
            unw_file_list=unw_files,
            reference=ref_point,
            output_dir=output_dir,
            # ifg_date_pairs: Sequence[Sequence[DateOrDatetime]] | None = None,
            # block_shape: tuple[int, int] = (512, 512),
            # cor_file_list: Sequence[PathOrStr] | None = None,
            # cor_threshold: float = 0.2,
            num_threads=1,
        )
        # Check results
        solved_stack = io.RasterStackReader.from_file_list(out_files)[:, :, :]
        sar_phases = data[1]
        npt.assert_allclose(solved_stack, sar_phases[1:], atol=1e-5)


class TestVelocity:
    @pytest.fixture
    def x_arr(self, data):
        sar_dates, sar_phases, ifg_date_pairs, ifgs = data
        return timeseries.datetime_to_float(sar_dates)

    @pytest.fixture
    def expected_velo(self, data, x_arr):
        sar_dates, sar_phases, ifg_date_pairs, ifgs = data
        out = np.zeros(SHAPE)
        for i in range(SHAPE[0]):
            for j in range(SHAPE[1]):
                out[i, j] = np.polyfit(x_arr, sar_phases[:, i, j], 1)[0]
        return out * 365.25

    def test_stack(self, data, x_arr, expected_velo):
        sar_dates, sar_phases, ifg_date_pairs, ifgs = data

        weights = np.ones_like(sar_phases)
        velocities = timeseries.estimate_velocity(x_arr, sar_phases, weights)
        assert velocities.shape == (SHAPE[0], SHAPE[1])
        npt.assert_allclose(velocities, expected_velo, atol=1e-5)

    def test_stack_unweighted(self, data, x_arr, expected_velo):
        sar_dates, sar_phases, ifg_date_pairs, ifgs = data

        velocities = timeseries.estimate_velocity(x_arr, sar_phases, None)
        assert velocities.shape == (SHAPE[0], SHAPE[1])
        npt.assert_allclose(velocities, expected_velo, atol=1e-5)


if __name__ == "__main__":
    sar_dates = make_sar_dates()
    sar_phases = make_sar_phases(sar_dates)
    ifg_date_pairs = make_ifg_date_pairs(sar_dates)
    ifgs = make_ifgs(sar_phases)



import os

import numpy as np
import numpy.testing as npt
import pytest

from dolphin._types import HalfWindow, Strides
from dolphin.phase_link import _compress, _core, simulate
from dolphin.utils import gpu_is_available

GPU_AVAILABLE = gpu_is_available() and os.environ.get("NUMBA_DISABLE_JIT") != "1"


@pytest.fixture
def slc_samples():
    C = np.eye(10)
    ns = 11 * 11
    return simulate.simulate_neighborhood_stack(C, ns)


@pytest.mark.parametrize("strides", [1, 3])
def test_compression(slc_samples, strides):
    slc_stack = slc_samples.reshape(10, 11, 11)

    # cpx_phase, temp_coh, eigs, _ = _core.run_cpl(
    pl_out = _core.run_cpl(
        slc_stack,
        HalfWindow(x=3, y=3),
        Strides(x=strides, y=strides),
    )
    comp_slc = _compress.compress(slc_stack=slc_stack, pl_cpx_phase=pl_out.cpx_phase)
    assert comp_slc.shape == slc_stack.shape[1:]

    # When striding, the upsampling will leave some nans at the edges
    valid_rows = slc_stack.shape[1] // strides * strides
    valid_cols = slc_stack.shape[2] // strides * strides

    stack_mean = np.mean(np.abs(slc_stack), axis=0)
    npt.assert_array_almost_equal(
        stack_mean[:valid_rows, :valid_cols], np.abs(comp_slc)[:valid_rows, :valid_cols]
    )
    assert np.isnan(np.abs(comp_slc)[valid_rows:, :]).all()
    assert np.isnan(np.abs(comp_slc)[:, valid_cols:]).all()



from pathlib import Path

import pytest
from make_netcdf import create_test_nc
from pyproj import CRS

from dolphin import io, stitching
from dolphin._types import Bbox


@pytest.fixture()
def shifted_slc_files(tmp_path):
    """Make series of files offset in lat/lon."""
    shape = (5, 10, 10)

    start_date = 20220101
    name_template = tmp_path / "shifted_{date}.nc"
    file_list = []
    for i in range(shape[0]):
        fname = str(name_template).format(date=str(start_date + i))
        create_test_nc(fname, epsg=4326, subdir="/", shape=shape[1:], xoff=i, yoff=i)
        file_list.append(Path(fname))

    return file_list


@pytest.fixture()
def shifted_slc_bounds():
    return Bbox(-5.5, -4.5, 8.5, 9.5)


def test_get_combined_bounds_gt(shifted_slc_files, shifted_slc_bounds):
    # Use the created WGS84 SLC
    bnds = io.get_raster_bounds(shifted_slc_files[0])
    expected_bnds = (-5.5, -4.5, 4.5, 5.5)
    assert bnds == expected_bnds

    bnds = io.get_raster_bounds(shifted_slc_files[1])
    expected_bnds = (-4.5, -3.5, 5.5, 6.5)
    assert bnds == expected_bnds

    bnds = io.get_raster_bounds(shifted_slc_files[-1])
    expected_bnds = (-1.5, -0.5, 8.5, 9.5)
    assert bnds == expected_bnds

    # should be same as test #1 above
    bnds, nd = stitching.get_combined_bounds_nodata(
        shifted_slc_files[0],
    )
    expected_bnds = (-5.5, -4.5, 4.5, 5.5)
    expected_nd = None
    assert bnds == expected_bnds
    assert nd == expected_nd

    # check same file twice
    bnds, nd = stitching.get_combined_bounds_nodata(
        shifted_slc_files[0], shifted_slc_files[0]
    )
    assert bnds == expected_bnds

    # Now combined one: should have the mins as the first two values,
    # and the last two values should be the idx=-1 file
    bnds, nd = stitching.get_combined_bounds_nodata(*shifted_slc_files)
    assert bnds == shifted_slc_bounds


def test_get_combined_bounds_gt_different_proj(
    slc_file_list_nc, slc_file_list_nc_wgs84
):
    bnds, gt = stitching.get_combined_bounds_nodata(*slc_file_list_nc)
    assert bnds == (-5.5, -2.0, 4.5, 3.0)

    with pytest.raises(ValueError):
        stitching.get_combined_bounds_nodata(
            slc_file_list_nc_wgs84[0], slc_file_list_nc[0]
        )


def test_get_mode_projection(slc_file_list_nc, slc_file_list_nc_wgs84):
    p = stitching._get_mode_projection(slc_file_list_nc_wgs84[0:1])

    epsg4326 = CRS.from_epsg(4326)
    assert CRS.from_user_input(p) == epsg4326

    p = stitching._get_mode_projection(
        slc_file_list_nc[0:1],
    )
    epsg32615 = CRS.from_epsg(32615)
    assert CRS.from_user_input(p) == epsg32615

    p = stitching._get_mode_projection(
        slc_file_list_nc[0:2] + slc_file_list_nc_wgs84[0:1],
    )
    assert CRS.from_user_input(p) == epsg32615

    p = stitching._get_mode_projection(
        slc_file_list_nc[0:1] + slc_file_list_nc_wgs84[0:2],
    )
    assert CRS.from_user_input(p) == epsg4326


def test_merge_images(tmp_path, shifted_slc_files, shifted_slc_bounds):
    outfile = tmp_path / "stitched.tif"
    stitching.merge_images(shifted_slc_files, outfile, target_aligned_pixels=False)

    b = io.get_raster_bounds(outfile)
    assert b == shifted_slc_bounds

    # The target aligned pixels option makes the bounds a multiple of dx/dy
    # which is 1.0/1.0 here
    outfile = tmp_path / "stitched_tap.tif"
    # Now the top/bottom must divide by 3, and left/right divide by 2
    stitching.merge_images(shifted_slc_files, outfile, target_aligned_pixels=True)

    b = io.get_raster_bounds(outfile)
    expected_bounds_tap = Bbox(-6.0, -5.0, 9.0, 10.0)
    assert b == expected_bounds_tap


def test_merge_images_strided(tmp_path, shifted_slc_files, shifted_slc_bounds):
    strides = {"x": 2, "y": 3}
    outfile = tmp_path / "stitched.tif"
    stitching.merge_images(
        shifted_slc_files,
        outfile,
        target_aligned_pixels=False,
        out_bounds=shifted_slc_bounds,
        strides=strides,
    )

    b = io.get_raster_bounds(outfile)
    assert b == shifted_slc_bounds

    # The target aligned pixels option makes the bounds a multiple of dx/dy
    # which is 1.0/1.0 here
    outfile = tmp_path / "stitched_tap.tif"
    expected_bounds_tap = Bbox(-6.0, -6.0, 10.0, 12.0)
    stitching.merge_images(
        shifted_slc_files, outfile, target_aligned_pixels=True, strides=strides
    )

    b = io.get_raster_bounds(outfile)
    assert b == expected_bounds_tap


@pytest.mark.parametrize("buffer", [0, 1.5])
@pytest.mark.parametrize("strides", [{"x": 1, "y": 1}, {"x": 3, "y": 2}])
def test_merge_images_specify_bounds(
    tmp_path, strides, buffer, shifted_slc_files, shifted_slc_bounds
):
    from shapely.geometry import box

    outfile = tmp_path / "stitched.tif"

    buffered_box = box(*shifted_slc_bounds).buffer(buffer)
    buffered_bounds = buffered_box.bounds
    stitching.merge_images(
        shifted_slc_files,
        outfile,
        target_aligned_pixels=False,
        out_bounds=buffered_bounds,
        strides=strides,
    )

    b = io.get_raster_bounds(outfile)
    assert b == buffered_bounds


def test_merge_images_one_image_out_bounds_specified(
    tmp_path, shifted_slc_files, shifted_slc_bounds
):
    from shapely.geometry import box

    outfile = tmp_path / "stitched.tif"

    buffered_box = box(*shifted_slc_bounds).buffer(1.5)
    buffered_bounds = buffered_box.bounds

    # only grab the first image
    # without specifying bounds, this should not work
    stitching.merge_images(shifted_slc_files[:1], outfile)

    b = io.get_raster_bounds(outfile)
    assert b != buffered_bounds

    outfile2 = tmp_path / "stitched_with_bounds.tif"
    # after specifying bounds, this should work
    stitching.merge_images(
        shifted_slc_files[:1],
        outfile2,
        target_aligned_pixels=False,
        out_bounds=buffered_bounds,
    )

    b = io.get_raster_bounds(outfile2)
    assert b == buffered_bounds



from pathlib import Path

import numpy as np
import pytest

from dolphin import filtering, io


def test_filter_long_wavelength():
    # Check filtering with ramp phase
    y, x = np.ogrid[-3:3:512j, -3:3:512j]
    unw_ifg = np.pi * (x + y)
    corr = np.ones(unw_ifg.shape, dtype=np.float32)
    bad_pixel_mask = corr < 0.5

    # Filtering
    filtered_ifg = filtering.filter_long_wavelength(
        unw_ifg, bad_pixel_mask=bad_pixel_mask, pixel_spacing=1000
    )
    np.testing.assert_allclose(
        filtered_ifg[10:-10, 10:-10],
        np.zeros(filtered_ifg[10:-10, 10:-10].shape),
        atol=1.0,
    )


def test_filter_long_wavelength_too_large_cutoff():
    # Check filtering with ramp phase
    y, x = np.ogrid[-3:3:512j, -3:3:512j]
    unw_ifg = np.pi * (x + y)
    bad_pixel_mask = np.zeros(unw_ifg.shape, dtype=bool)

    with pytest.raises(ValueError):
        filtering.filter_long_wavelength(
            unw_ifg,
            bad_pixel_mask=bad_pixel_mask,
            pixel_spacing=1,
            wavelength_cutoff=50_000,
        )


@pytest.fixture()
def unw_files(tmp_path):
    """Make series of files offset in lat/lon."""
    shape = (3, 9, 9)

    y, x = np.ogrid[-3:3:512j, -3:3:512j]
    file_list = []
    for i in range(shape[0]):
        unw_arr = (i + 1) * np.pi * (x + y)
        fname = tmp_path / f"unw_{i}.tif"
        io.write_arr(arr=unw_arr, output_name=fname)
        file_list.append(Path(fname))

    return file_list


def test_filter(tmp_path, unw_files):
    output_dir = Path(tmp_path) / "filtered"
    filtering.filter_rasters(
        unw_filenames=unw_files,
        output_dir=output_dir,
        max_workers=1,
        wavelength_cutoff=50,
    )



"""Utility methods to print system info for debugging.

Adapted from `rasterio.show_versions`,
which was adapted from `sklearn.utils._show_versions`
which was adapted from `pandas.show_versions`
"""

from __future__ import annotations

import importlib
import platform
import re
import sys
from importlib.metadata import metadata
from typing import Optional

import dolphin

__all__ = ["show_versions"]


def _get_sys_info() -> dict[str, str]:
    """System information.

    Returns
    -------
    dict
        system and Python version information

    """
    return {
        "python": sys.version.replace("\n", " "),
        "executable": sys.executable,
        "machine": platform.platform(),
    }


def _get_version(module_name: str) -> Optional[str]:
    if module_name in sys.modules:
        mod = sys.modules[module_name]
    else:
        try:
            mod = importlib.import_module(module_name.replace("-", "_"))
        except ImportError:
            return None
    try:
        return mod.__version__
    except AttributeError:
        return mod.version


def _get_unwrapping_options() -> dict[str, Optional[str]]:
    """Information on possible phase unwrapping libraries.

    Returns
    -------
    dict
        module information

    """
    return {
        "snaphu": _get_version("snaphu"),
        "spurt": _get_version("spurt"),
        "isce3": _get_version("isce3"),
        "tophu": _get_version("tophu"),
        "whirlwind": _get_version("whirlwind"),
    }


def _get_deps_info() -> dict[str, Optional[str]]:
    """Overview of the installed version of main dependencies.

    Returns
    -------
    dict:
        version information on relevant Python libraries

    """
    # Get metadata for your package
    meta = metadata("dolphin")
    # Extract dependencies from 'Requires-Dist' field
    deps = [
        re.split(r"[><=~!]", dep.split()[0])[0]
        for dep in meta.get_all("Requires-Dist", [])
        if "extra" not in dep
    ]
    # Replace 'ruamel-yaml' with 'ruamel.yaml'
    deps = [dep.replace("ruamel-yaml", "ruamel.yaml") for dep in deps]
    # Add `osgeo` for gdal (not listed in pip requirements)
    deps += ["osgeo.gdal"]
    return {name: _get_version(name) for name in deps}


def _get_gpu_info() -> dict[str, Optional[str]]:
    """Overview of the optional GPU packages.

    Returns
    -------
    dict:
        version information on relevant Python libraries

    """
    from dolphin.utils import gpu_is_available

    return {"jax": _get_version("jax"), "gpu_is_available": str(gpu_is_available())}


def _print_info_dict(info_dict: dict) -> None:
    """Print the information dictionary."""
    for key, stat in info_dict.items():
        print(f"{key:>12}: {stat}")


def show_versions() -> None:
    """Print useful debugging information.

    Examples
    --------
    > python -c "import dolphin; dolphin.show_versions()"

    """
    print(f"dolphin version: {dolphin.__version__}")
    print("\nPython deps:")
    _print_info_dict(_get_deps_info())
    print("\nSystem:")
    _print_info_dict(_get_sys_info())
    print("optional GPU info:")
    _print_info_dict(_get_gpu_info())



from __future__ import annotations

import logging
import shutil
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Callable, Optional, Protocol, Sequence, TypeVar

import jax.numpy as jnp
import numpy as np
from jax import Array, jit, vmap
from numpy.typing import ArrayLike
from opera_utils import get_dates
from scipy import ndimage

from dolphin import DateOrDatetime, io, utils
from dolphin._overviews import ImageType, create_overviews
from dolphin._types import PathOrStr, ReferencePoint
from dolphin.utils import flatten, format_dates, full_suffix
from dolphin.workflows import CallFunc

T = TypeVar("T")

logger = logging.getLogger(__name__)

__all__ = ["run"]


class ReferencePointError(ValueError):
    pass


def run(
    unwrapped_paths: Sequence[PathOrStr],
    conncomp_paths: Sequence[PathOrStr],
    corr_paths: Sequence[PathOrStr],
    condition_file: PathOrStr,
    condition: CallFunc,
    output_dir: PathOrStr,
    run_velocity: bool = False,
    velocity_file: Optional[PathOrStr] = None,
    correlation_threshold: float = 0.2,
    num_threads: int = 5,
    reference_point: tuple[int, int] = (-1, -1),
) -> tuple[list[Path], ReferencePoint]:
    """Invert the unwrapped interferograms, estimate timeseries and phase velocity.

    Parameters
    ----------
    unwrapped_paths : Sequence[Path]
        Sequence unwrapped interferograms to invert.
    corr_paths : Sequence[Path]
        Sequence interferometric correlation files, one per file in `unwrapped_paths`
    conncomp_paths : Sequence[Path]
        Sequence connected component files, one per file in `unwrapped_paths`
    condition_file: PathOrStr
        A file with the same size as each raster, like amplitude dispersion or
        temporal coherence
    condition: CallFunc
        The function to apply to the condition file,
        for example numpy.argmin which finds the pixel with lowest value
        the options are [min, max]
    output_dir : Path
        Path to the output directory.
    run_velocity : bool
        Whether to run velocity estimation on the inverted phase series
    velocity_file : Path, Optional
        The output velocity file
    correlation_threshold : float
        Pixels with correlation below this value will be masked out
    num_threads : int
        The parallel blocks to process at once.
        Default is 5.
    reference_point : tuple[int, int], optional
        Reference point (row, col) used if performing a time series inversion.
        If not provided, a point will be selected from a consistent connected
        component with low amplitude dispersion or high temporal coherence.

    Returns
    -------
    inverted_phase_paths : list[Path]
        list of Paths to inverted interferograms (single reference phase series).
    reference_point : ReferencePoint
        NamedTuple of reference (row, column) selected.
        If passed as input, simply returned back as output.
        Otherwise, the result is the auto-selection from `select_reference_point`.

    """
    Path(output_dir).mkdir(exist_ok=True, parents=True)

    condition_func = argmax_index if condition == CallFunc.MAX else argmin_index
    if reference_point == (-1, -1):
        ref_point = select_reference_point(
            condition_file=condition_file,
            output_dir=Path(output_dir),
            condition_func=condition_func,
            ccl_file_list=conncomp_paths,
        )
    else:
        ref_point = ReferencePoint(row=reference_point[0], col=reference_point[1])

    ifg_date_pairs = [get_dates(f) for f in unwrapped_paths]
    sar_dates = sorted(set(utils.flatten(ifg_date_pairs)))
    # if we did single-reference interferograms, for `n` sar dates, we will only have
    # `n-1` interferograms. Any more than n-1 ifgs means we need to invert
    needs_inversion = len(unwrapped_paths) > len(sar_dates) - 1
    # check if we even need to invert, or if it was single reference
    inverted_phase_paths: list[Path] = []
    if needs_inversion:
        logger.info("Selecting a reference point for unwrapped interferograms")

        logger.info("Inverting network of %s unwrapped ifgs", len(unwrapped_paths))
        inverted_phase_paths = invert_unw_network(
            unw_file_list=unwrapped_paths,
            reference=ref_point,
            output_dir=output_dir,
            num_threads=num_threads,
        )
    else:
        logger.info(
            "Skipping inversion step: only single reference interferograms exist."
        )
        # Copy over the unwrapped paths to `timeseries/`
        for p in unwrapped_paths:
            # if it ends in `.unw.tif`, change to just `.tif` for consistency
            # with the case where we run an inversion
            cur_name = Path(p).name
            unw_suffix = full_suffix(p)
            target_name = str(cur_name).replace(unw_suffix, ".tif")
            target = Path(output_dir) / target_name
            if not target.exists():  # Check to prevent overwriting
                shutil.copy(p, target)
            inverted_phase_paths.append(target)

    if run_velocity:
        #  We can't pass the correlations after an inversion- the numbers don't match
        # TODO:
        # Is there a better weighting then?
        cor_file_list = (
            corr_paths if len(corr_paths) == len(inverted_phase_paths) else None
        )
        logger.info("Estimating phase velocity")
        if velocity_file is None:
            velocity_file = Path(output_dir) / "velocity.tif"

        create_velocity(
            unw_file_list=inverted_phase_paths,
            output_file=velocity_file,
            reference=ref_point,
            cor_file_list=cor_file_list,
            cor_threshold=correlation_threshold,
            num_threads=num_threads,
        )

    return inverted_phase_paths, ref_point


def argmin_index(arr: ArrayLike) -> tuple[int, ...]:
    """Get the index tuple of the minimum value of the array.

    If multiple occurrences of the minimum value exist, returns
    the index of the first such occurrence in the flattened array.

    Parameters
    ----------
    arr : array_like
        The input array.

    Returns
    -------
    tuple of int
        The index of the minimum value.

    """
    return np.unravel_index(np.argmin(arr), np.shape(arr))


def argmax_index(arr: ArrayLike) -> tuple[int, ...]:
    """Get the index tuple of the maximum value of the array.

    If multiple occurrences of the maximum value exist, returns
    the index of the first such occurrence in the flattened array.

    Parameters
    ----------
    arr : array_like
        The input array.

    Returns
    -------
    tuple of int
        The index of the maximum value.

    """
    return np.unravel_index(np.argmax(arr), np.shape(arr))


@jit
def weighted_lstsq_single(
    A: ArrayLike,
    b: ArrayLike,
    weights: ArrayLike,
) -> Array:
    r"""Perform weighted least for one data vector.

    Minimizes the weighted 2-norm of the residual vector:

    \[
        || b - A x ||^2_W
    \]

    where \(W\) is a diagonal matrix of weights.

    Parameters
    ----------
    A : Arraylike
        Incidence matrix of shape (n_ifgs, n_sar_dates - 1)
    b : ArrayLike, 1D
        The phase differences between the ifg pairs
    weights : ArrayLike, 1D, optional
        The weights for each element of `b`.

    Returns
    -------
    x : np.array 1D
        The estimated phase for each SAR acquisition
    residuals : np.array 1D
        Sums of squared residuals: Squared Euclidean 2-norm for `b - A @ x`
        For a 1D `b`, this is a scalar.

    """
    # scale both A and b by sqrt so we are minimizing
    sqrt_weights = jnp.sqrt(weights)
    # Multiply each data point by sqrt(weight)
    b_scaled = b * sqrt_weights
    # Multiply each row by sqrt(weight)
    A_scaled = A * sqrt_weights[:, None]

    # Run the weighted least squares
    x, residuals, rank, sing_vals = jnp.linalg.lstsq(A_scaled, b_scaled)
    # TODO: do we need special handling?
    # if rank < A.shape[1]:
    #     # logger.warning("Rank deficient solution")

    return x, residuals.ravel()


@jit
def invert_stack(
    A: ArrayLike, dphi: ArrayLike, weights: ArrayLike | None = None
) -> Array:
    """Solve the SBAS problem for a stack of unwrapped phase differences.

    Parameters
    ----------
    A : ArrayLike
        Incidence matrix of shape (n_ifgs, n_sar_dates - 1)
    dphi : ArrayLike
        The phase differences between the ifg pairs, shape=(n_ifgs, n_rows, n_cols)
    weights : ArrayLike, optional
        The weights for each element of `dphi`.
        Same shape as `dphi`.
        If not provided, all weights are set to 1 (ordinary least squares).

    Returns
    -------
    phi : np.array 3D
        The estimated phase for each SAR acquisition
        Shape is (n_sar_dates - 1, n_rows, n_cols)
    residuals : np.array 2D
        Sums of squared residuals: Squared Euclidean 2-norm for `dphi - A @ x`
        Shape is (n_rows, n_cols)

    Notes
    -----
    To mask out data points of a pixel, the weight can be set to 0.
    When `A` remains full rank, setting the weight to zero is the same as removing
    the entry from the data vector and the corresponding row from `A`.

    """
    n_ifgs, n_rows, n_cols = dphi.shape

    if weights is None:
        # Can use ordinary least squares with no weights
        # Reshape to be size (M, K) instead of 3D
        b = dphi.reshape(n_ifgs, -1)
        phase_cols, residuals_cols, _, _ = jnp.linalg.lstsq(A, b)
        # Reshape the phase and residuals to be 3D
        phase = phase_cols.reshape(-1, n_rows, n_cols)
        residuals = residuals_cols.reshape(n_rows, n_cols)
    else:
        # vectorize the solve function to work on 2D and 3D arrays
        # We are not vectorizing over the A matrix, only the dphi vector
        # Solve 2d shapes: (nrows, n_ifgs) -> (nrows, n_sar_dates)
        invert_2d = vmap(weighted_lstsq_single, in_axes=(None, 1, 1), out_axes=(1, 1))
        # Solve 3d shapes: (nrows, ncols, n_ifgs) -> (nrows, ncols, n_sar_dates)
        invert_3d = vmap(invert_2d, in_axes=(None, 2, 2), out_axes=(2, 2))
        phase, residuals = invert_3d(A, dphi, weights)
        # Reshape the residuals to be 2D
        residuals = residuals[0]

    return phase, residuals


def get_incidence_matrix(
    ifg_pairs: Sequence[tuple[T, T]], sar_idxs: Sequence[T] | None = None
) -> np.ndarray:
    """Build the indicator matrix from a list of ifg pairs (index 1, index 2).

    Parameters
    ----------
    ifg_pairs : Sequence[tuple[T, T]]
        List of ifg pairs represented as tuples of (day 1, day 2)
        Can be ints, datetimes, etc.
    sar_idxs : Sequence[T], optional
        If provided, used as the total set of indexes which `ifg_pairs`
        were formed from.
        Otherwise, created from the unique entries in `ifg_pairs`.
        Only provide if there are some dates which are not present in `ifg_pairs`.

    Returns
    -------
    A : np.array 2D
        The incident-like matrix for the system: A*phi = dphi
        Each row corresponds to an ifg, each column to a SAR date.
        The value will be -1 on the early (reference) ifgs, +1 on later (secondary)
        since the ifg phase = (later - earlier)
        Shape: (n_ifgs, n_sar_dates - 1)

    """
    if sar_idxs is None:
        sar_idxs = sorted(set(flatten(ifg_pairs)))

    M = len(ifg_pairs)
    N = len(sar_idxs) - 1
    A = np.zeros((M, N))

    # Create a dictionary mapping sar dates to matrix columns
    # We take the first SAR acquisition to be time 0, leave out of matrix
    date_to_col = {date: i for i, date in enumerate(sar_idxs[1:])}
    # Populate the matrix
    for i, (early, later) in enumerate(ifg_pairs):
        if early in date_to_col:
            A[i, date_to_col[early]] = -1
        if later in date_to_col:
            A[i, date_to_col[later]] = +1

    return A


@jit
def estimate_velocity_pixel(x: ArrayLike, y: ArrayLike, w: ArrayLike) -> Array:
    """Estimate the velocity from a single pixel's time series.

    Parameters
    ----------
    x : np.array 1D
        The time values
    y : np.array 1D
        The unwrapped phase values
    w : np.array 1D
        The weights for each time value

    Returns
    -------
    velocity : np.array, 0D
        The estimated velocity in (unw unit) / year.

    """
    # Jax polyfit will grab the first *2* dimensions of y to solve in a batch
    return jnp.polyfit(x, y, deg=1, w=w.reshape(y.shape), rcond=None)[0]


@jit
def estimate_velocity(
    x_arr: ArrayLike, unw_stack: ArrayLike, weight_stack: ArrayLike | None
) -> Array:
    """Estimate the velocity from a stack of unwrapped interferograms.

    Parameters
    ----------
    x_arr : ArrayLike
        Array of time values corresponding to each unwrapped phase image.
        Length must match `unw_stack.shape[0]`.
    unw_stack : ArrayLike
        Array of unwrapped phase values at each pixel, shape=`(n_time, n_rows, n_cols)`.
    weight_stack : ArrayLike, optional
        Array of weights for each pixel, shape=`(n_time, n_rows, n_cols)`.
        If not provided, performs one batch unweighted linear fit.

    Returns
    -------
    velocity : np.array 2D
        The estimated velocity in (unw unit) / year calculated as 365.25 * rad/day.
        E.g. if the unwrapped phase is in radians, the velocity is in rad/year.

    """
    # TODO: weighted least squares using correlation?
    n_time, n_rows, n_cols = unw_stack.shape

    unw_pixels = unw_stack.reshape(n_time, -1)
    if weight_stack is None:
        # For jnp.polyfit(...), coeffs[0] is slope, coeffs[1] is the intercept
        velos = jnp.polyfit(x_arr, unw_pixels, deg=1, rcond=None)[0]
    else:
        # We use the same x inputs for all output pixels
        if unw_stack.shape != weight_stack.shape:
            msg = (
                "unw_stack and weight_stack must have the same shape,"
                f" got {unw_stack.shape} and {weight_stack.shape}"
            )
            raise ValueError(msg)

        weights_pixels = weight_stack.reshape(n_time, 1, -1)

        velos = vmap(estimate_velocity_pixel, in_axes=(None, -1, -1))(
            x_arr, unw_pixels, weights_pixels
        )
    # Currently `velos` is in units / day,
    days_per_year = 365.25
    return velos.reshape(n_rows, n_cols) * days_per_year


def datetime_to_float(dates: Sequence[DateOrDatetime]) -> np.ndarray:
    """Convert a sequence of datetime objects to a float representation.

    Output units are in days since the first item in `dates`.

    Parameters
    ----------
    dates : Sequence[DateOrDatetime]
        List of datetime objects to convert to floats

    Returns
    -------
    date_arr : np.array 1D
        The float representation of the datetime objects

    """
    sec_per_day = 60 * 60 * 24
    date_arr = np.asarray(dates).astype("datetime64[s]")
    # Reference the 0 to the first date
    date_arr = date_arr - date_arr[0]
    return date_arr.astype(float) / sec_per_day


def create_velocity(
    unw_file_list: Sequence[PathOrStr],
    output_file: PathOrStr,
    reference: ReferencePoint,
    date_list: Sequence[DateOrDatetime] | None = None,
    cor_file_list: Sequence[PathOrStr] | None = None,
    cor_threshold: float = 0.2,
    block_shape: tuple[int, int] = (512, 512),
    num_threads: int = 5,
    add_overviews: bool = True,
) -> None:
    """Perform pixel-wise (weighted) linear regression to estimate velocity.

    The units of `output_file` are in (unwrapped units) / year.
    E.g. if the files in `unw_file_list` are in radians, the output velocity
    is in radians / year, which is calculated as 365.25 * radians / day.

    Parameters
    ----------
    unw_file_list : Sequence[PathOrStr]
        List of unwrapped phase files.
    output_file : PathOrStr
        The output file to save the velocity to.
    reference : ReferencePoint
        The (row, col) to use as reference before fitting the velocity.
        This point will be subtracted from all other points before solving.
    date_list : Sequence[DateOrDatetime], optional
        List of dates corresponding to the unwrapped phase files.
        If not provided, will be parsed from filenames in `unw_file_list`.
    cor_file_list : Sequence[PathOrStr], optional
        List of correlation files to use for weighting the velocity estimation.
        If not provided, all weights are set to 1.
    cor_threshold : float, optional
        The correlation threshold to use for weighting the velocity estimation.
        Default is 0.2.
    block_shape : tuple[int, int], optional
        The shape of the blocks to process in parallel.
        Default is (512, 512)
    num_threads : int, optional
        The parallel blocks to process at once.
        Default is 5.
    add_overviews : bool, optional
        If True, creates overviews of the new velocity raster.
        Default is True.

    """
    if Path(output_file).exists():
        logger.info(f"Output file {output_file} already exists, skipping velocity")
        return

    if date_list is None:
        date_list = [get_dates(f)[1] for f in unw_file_list]
    x_arr = datetime_to_float(date_list)

    # Set up the input data readers
    out_dir = Path(output_file).parent
    unw_reader = io.VRTStack(
        file_list=unw_file_list,
        outfile=out_dir / "velocity_inputs.vrt",
        skip_size_check=True,
    )
    if cor_file_list is not None:
        if len(cor_file_list) != len(unw_file_list):
            msg = "Mismatch in number of input files provided:"
            msg += f"{len(cor_file_list) = }, but {len(unw_file_list) = }"
            raise ValueError(msg)

        logger.info("Using correlation to weight velocity fit")
        cor_reader = io.VRTStack(
            file_list=cor_file_list,
            outfile=out_dir / "cor_inputs.vrt",
            skip_size_check=True,
        )
    else:
        logger.info("Using unweighted fit for velocity.")
        cor_reader = None

    # Read in the reference point
    ref_row, ref_col = reference
    logger.info(f"Reading phase reference pixel {reference}")
    ref_data = unw_reader[:, ref_row, ref_col].reshape(-1, 1, 1)

    def read_and_fit(
        readers: Sequence[io.StackReader], rows: slice, cols: slice
    ) -> tuple[np.ndarray, slice, slice]:
        # Only use the cor_reader if it's the same shape as the unw_reader
        if len(readers) == 2:
            unw_reader, cor_reader = readers
            unw_stack = unw_reader[:, rows, cols]
            weights = cor_reader[:, rows, cols]
            weights[weights < cor_threshold] = 0
        else:
            unw_stack = readers[0][:, rows, cols]
            # weights = np.ones_like(unw_stack)
            weights = None
        # Reference the data
        unw_stack = unw_stack - ref_data
        # Fit a line to each pixel with weighted least squares
        return (
            estimate_velocity(x_arr=x_arr, unw_stack=unw_stack, weight_stack=weights),
            rows,
            cols,
        )

    # Note: For some reason, the `RasterStackReader` is much slower than the VRT
    # for files on S3:
    # ~300 files takes >2 min to open, >2 min to read each block
    # VRTStack seems to take ~30 secs to open, 1 min to read
    # Very possible there's a tuning param/rasterio config to fix, but not sure.
    readers = [unw_reader]
    if cor_reader is not None:
        readers.append(cor_reader)

    writer = io.BackgroundRasterWriter(output_file, like_filename=unw_file_list[0])
    io.process_blocks(
        readers=readers,
        writer=writer,
        func=read_and_fit,
        block_shape=block_shape,
        num_threads=num_threads,
    )

    writer.notify_finished()
    if add_overviews:
        logger.info("Creating overviews for velocity image")
        create_overviews([output_file])
    logger.info("Completed create_velocity")


class AverageFunc(Protocol):
    """Protocol for temporally averaging a block of data."""

    def __call__(self, ArrayLike, axis: int) -> ArrayLike: ...


def create_temporal_average(
    file_list: Sequence[PathOrStr],
    output_file: PathOrStr,
    block_shape: tuple[int, int] = (512, 512),
    num_threads: int = 5,
    average_func: Callable[[ArrayLike, int], np.ndarray] = np.nanmean,
    read_masked: bool = False,
) -> None:
    """Average all images in `reader` to create a 2D image in `output_file`.

    Parameters
    ----------
    file_list : Sequence[PathOrStr]
        List of files to average
    output_file : PathOrStr
        The output file to save the average to
    block_shape : tuple[int, int], optional
        The shape of the blocks to process in parallel.
        Default is (512, 512)
    num_threads : int, optional
        The parallel blocks to process at once.
        Default is 5.
    average_func : Callable[[ArrayLike, int], np.ndarray], optional
        The function to use to average the images.
        Default is `np.nanmean`, which calls `np.nanmean(arr, axis=0)` on each block.
    read_masked : bool, optional
        If True, reads the data as a masked array based on the rasters' nodata values.
        Default is False.

    """

    def read_and_average(
        readers: Sequence[io.StackReader], rows: slice, cols: slice
    ) -> tuple[slice, slice, np.ndarray]:
        chunk = readers[0][:, rows, cols]
        return average_func(chunk, 0), rows, cols

    writer = io.BackgroundRasterWriter(output_file, like_filename=file_list[0])
    with NamedTemporaryFile(mode="w", suffix=".vrt") as f:
        reader = io.VRTStack(
            file_list=file_list,
            outfile=f.name,
            skip_size_check=True,
            read_masked=read_masked,
        )

        io.process_blocks(
            readers=[reader],
            writer=writer,
            func=read_and_average,
            block_shape=block_shape,
            num_threads=num_threads,
        )

    writer.notify_finished()


def invert_unw_network(
    unw_file_list: Sequence[PathOrStr],
    reference: ReferencePoint,
    output_dir: PathOrStr,
    cor_file_list: Sequence[PathOrStr] | None = None,
    cor_threshold: float = 0.2,
    n_cor_looks: int = 1,
    ifg_date_pairs: Sequence[Sequence[DateOrDatetime]] | None = None,
    block_shape: tuple[int, int] = (512, 512),
    num_threads: int = 5,
    add_overviews: bool = True,
) -> list[Path]:
    """Perform pixel-wise inversion of unwrapped network to get phase per date.

    Parameters
    ----------
    unw_file_list : Sequence[PathOrStr]
        List of unwrapped phase files.
    reference : ReferencePoint
        The reference point to use for the inversion.
        The data vector from `unw_file_list` at this point will be subtracted
        from all other points when solving.
    output_dir : PathOrStr
        The directory to save the output files
    ifg_date_pairs : Sequence[Sequence[DateOrDatetime]], optional
        List of date pairs to use for the inversion. If not provided, will be
        parsed from filenames in `unw_file_list`.
    block_shape : tuple[int, int], optional
        The shape of the blocks to process in parallel
    cor_file_list : Sequence[PathOrStr], optional
        List of correlation files to use for weighting the inversion
    cor_threshold : float, optional
        The correlation threshold to use for weighting the inversion
        Default is 0.2
    n_cor_looks : int, optional
        The number of looks used to form the input correlation data, used
        to convert correlation to phase variance.
        Default is 1.
    num_threads : int
        The parallel blocks to process at once.
        Default is 5.
    add_overviews : bool, optional
        If True, creates overviews of the new unwrapped phase rasters.
        Default is True.

    Returns
    -------
    out_paths : list[Path]
        List of the output files created by the inversion.

    """
    if ifg_date_pairs is None:
        ifg_date_pairs = [get_dates(f) for f in unw_file_list]

    try:
        # Ensure it's a list of pairs
        ifg_tuples = [(ref, sec) for (ref, sec) in ifg_date_pairs]  # noqa: C416
    except ValueError as e:
        raise ValueError(
            "Each item in `ifg_date_pairs` must be a sequence of length 2"
        ) from e

    # Make the names of the output files from the SAR dates to solve for
    sar_dates = sorted(set(flatten(ifg_tuples)))
    ref_date = sar_dates[0]
    suffix = ".tif"
    # Create the `n_sar_dates - 1` output files (skipping the 0 reference raster)
    out_paths = [
        Path(output_dir) / (format_dates(ref_date, d) + suffix) for d in sar_dates[1:]
    ]
    if all(p.exists() for p in out_paths):
        logger.info("All output files already exist, skipping inversion")
        return out_paths

    A = get_incidence_matrix(ifg_pairs=ifg_tuples, sar_idxs=sar_dates)

    out_vrt_name = Path(output_dir) / "unw_network.vrt"
    unw_reader = io.VRTStack(
        file_list=unw_file_list, outfile=out_vrt_name, skip_size_check=True
    )
    cor_vrt_name = Path(output_dir) / "cor_network.vrt"

    # Get the reference point data
    ref_row, ref_col = reference
    ref_data = unw_reader[:, ref_row, ref_col].reshape(-1, 1, 1)

    def read_and_solve(
        readers: Sequence[io.StackReader], rows: slice, cols: slice
    ) -> tuple[slice, slice, np.ndarray]:
        if len(readers) == 2:
            unw_reader, cor_reader = readers
            stack = unw_reader[:, rows, cols]
            cor = cor_reader[:, rows, cols]
            weights = correlation_to_variance(cor, n_cor_looks)
            weights[cor < cor_threshold] = 0
        else:
            stack = readers[0][:, rows, cols]
            weights = None

        # subtract the reference
        stack = stack - ref_data

        # TODO: possible second input for weights? from conncomps
        # TODO: do i want to write residuals too? Do i need
        # to have multiple writers then?
        phases = invert_stack(A, stack, weights)[0]
        return np.asarray(phases), rows, cols

    if cor_file_list is not None:
        cor_reader = io.VRTStack(
            file_list=cor_file_list, outfile=cor_vrt_name, skip_size_check=True
        )
        readers = [unw_reader, cor_reader]
    else:
        readers = [unw_reader]

    writer = io.BackgroundStackWriter(out_paths, like_filename=unw_file_list[0])

    io.process_blocks(
        readers=readers,
        writer=writer,
        func=read_and_solve,
        block_shape=block_shape,
        num_threads=num_threads,
    )
    writer.notify_finished()

    if add_overviews:
        logger.info("Creating overviews for unwrapped images")
        create_overviews(out_paths, image_type=ImageType.UNWRAPPED)

    logger.info("Completed invert_unw_network")
    return out_paths


def correlation_to_variance(correlation: ArrayLike, nlooks: int) -> Array:
    r"""Convert interferometric correlation to phase variance.

    Uses the CRLB formula from Rodriguez, 1992 [1]_ to get the phase variance,
    \sigma_{\phi}^2:

    \[
        \sigma_{\phi}^{2} = \frac{1}{2N_{L}} \frac{1 - \gamma^{2}}{\gamma^{2}}
    \]

    where \gamma is the correlation and N_L is the effective number of looks.

    References
    ----------
    .. [1] Rodriguez, E., and J. M. Martin. "Theory and design of interferometric
    synthetic aperture radars." IEEE Proceedings of Radar and Signal Processing.
    Vol. 139. No. 2. IET Digital Library, 1992.

    """
    return (1 - correlation**2) / (2 * nlooks * correlation**2 + 1e-6)


def select_reference_point(
    *,
    condition_file: PathOrStr,
    output_dir: Path,
    condition_func: Callable[[ArrayLike], tuple[int, ...]] = argmin_index,
    ccl_file_list: Sequence[PathOrStr] | None = None,
    block_shape: tuple[int, int] = (512, 512),
    num_threads: int = 5,
) -> ReferencePoint:
    """Automatically select a reference point for a stack of unwrapped interferograms.

    Uses the condition file and (optionally) connected component labels.
    The point is selected which

    1. has the condition applied to condition file. for example: has the lowest
       amplitude dispersion
    2. (optionally) is within intersection of all nonzero connected component labels

    Parameters
    ----------
    condition_file: PathOrStr
        A file with the same size as each raster, like amplitude dispersion or
        temporal coherence in `ccl_file_list`
    output_dir: Path
        Path to store the computed "conncomp_intersection.tif" raster
    condition_func: Callable[[ArrayLike, ]]
        The function to apply to the condition file,
        for example numpy.argmin which finds the pixel with lowest value
    ccl_file_list : Sequence[PathOrStr]
        List of connected component label phase files.
    block_shape: tuple[int, int]
        Size of blocks to read from while processing `ccl_file_list`
        Default = (512, 512)
    num_threads: int
        Number of parallel blocks to process.
        Default = 5

    Returns
    -------
    ReferencePoint
        The select (row, col) as a namedtuple

    Raises
    ------
    ReferencePointError
        Raised if no valid region is found in the intersection of the connected
        component label files

    """
    output_file = output_dir / "reference_point.txt"
    if output_file.exists():
        ref_point = _read_reference_point(output_file=output_file)
        logger.info(f"Read {ref_point!r} from existing {output_file}")
        return ref_point

    logger.info("Selecting reference point")
    condition_file_values = io.load_gdal(condition_file, masked=True)

    isin_largest_conncomp = np.ones(condition_file_values.shape, dtype=bool)
    if ccl_file_list:
        try:
            isin_largest_conncomp = _get_largest_conncomp_mask(
                ccl_file_list=ccl_file_list,
                output_dir=output_dir,
                block_shape=block_shape,
                num_threads=num_threads,
            )
        except ReferencePointError:
            msg = "Unable to find find a connected component intersection."
            msg += f"Proceeding using only {condition_file = }"
            logger.warning(msg, exc_info=True)

    # Mask out where the conncomps aren't equal to the largest
    condition_file_values.mask = condition_file_values.mask | (~isin_largest_conncomp)

    # Pick the (unmasked) point with the condition applied to condition file
    ref_row, ref_col = condition_func(condition_file_values)

    # Cast to `int` to avoid having `np.int64` types
    ref_point = ReferencePoint(int(ref_row), int(ref_col))
    logger.info(f"Saving {ref_point!r} to from existing {output_file}")
    _write_reference_point(output_file=output_file, ref_point=ref_point)
    return ref_point


def _write_reference_point(output_file: Path, ref_point: ReferencePoint) -> None:
    output_file.write_text(",".join(list(map(str, ref_point))))


def _read_reference_point(output_file: Path):
    return ReferencePoint(*[int(n) for n in output_file.read_text().split(",")])


def _get_largest_conncomp_mask(
    output_dir: Path,
    ccl_file_list: Sequence[PathOrStr] | None = None,
    block_shape: tuple[int, int] = (512, 512),
    num_threads: int = 5,
) -> np.ndarray:
    def intersect_conncomp(arr: np.ma.MaskedArray, axis: int) -> np.ndarray:
        # Track where input is nodata
        any_masked = np.any(arr.mask, axis=axis)
        # Get the logical AND of all nonzero conncomp labels
        fillval = arr.fill_value
        is_valid_conncomp = arr.filled(0) > 0
        all_are_valid = np.all(is_valid_conncomp, axis=axis).astype(arr.dtype)
        # Reset nodata
        all_are_valid[any_masked] = fillval
        return all_are_valid

    conncomp_intersection_file = Path(output_dir) / "conncomp_intersection.tif"
    if ccl_file_list and not conncomp_intersection_file.exists():
        logger.info("Creating intersection of connected components")
        create_temporal_average(
            file_list=ccl_file_list,
            output_file=conncomp_intersection_file,
            block_shape=block_shape,
            num_threads=num_threads,
            average_func=intersect_conncomp,
            read_masked=True,
        )

    conncomp_intersection = io.load_gdal(conncomp_intersection_file, masked=True)

    # Find the largest conncomp region in the intersection
    label, nlabels = ndimage.label(
        conncomp_intersection.filled(0), structure=np.ones((3, 3))
    )
    if nlabels == 0:
        raise ReferencePointError(
            "Connected components intersection left no valid regions"
        )
    logger.info("Found %d connected components in intersection", nlabels)

    # Make a mask of the largest conncomp:
    # Find the label with the most pixels using bincount
    label_counts = np.bincount(label.ravel())
    # (ignore the 0 label)
    largest_idx = np.argmax(label_counts[1:]) + 1
    # Create a mask of pixels with this label
    isin_largest_conncomp = label == largest_idx
    return isin_largest_conncomp



from __future__ import annotations

from pathlib import Path

import pytest
from opera_utils import group_by_burst

from dolphin.utils import flatten, full_suffix
from dolphin.workflows import config, displacement

pytestmark = pytest.mark.filterwarnings(
    "ignore::rasterio.errors.NotGeoreferencedWarning",
    "ignore:.*io.FileIO.*:pytest.PytestUnraisableExceptionWarning",
)


def test_displacement_run_single(
    opera_slc_files: list[Path],
    tmpdir,
):
    with tmpdir.as_cwd():
        cfg = config.DisplacementWorkflow(
            cslc_file_list=opera_slc_files,
            input_options={"subdataset": "/data/VV"},
            interferogram_network={
                "indexes": [(0, -1)],
                "max_bandwidth": 2,
            },
            phase_linking={
                "ministack_size": 500,
            },
        )
        paths = displacement.run(cfg)

        for slc_paths in paths.comp_slc_dict.values():
            assert all(p.exists() for p in slc_paths)
        assert paths.stitched_ps_file.exists()
        assert all(p.exists() for p in paths.stitched_ifg_paths)
        assert all(p.exists() for p in paths.stitched_cor_paths)
        assert paths.stitched_temp_coh_file.exists()
        assert paths.stitched_ps_file.exists()
        assert paths.stitched_amp_dispersion_file.exists()
        assert paths.unwrapped_paths is not None
        assert paths.conncomp_paths is not None
        assert paths.timeseries_paths is not None
        assert all(p.exists() for p in paths.conncomp_paths)
        assert all(p.exists() for p in paths.unwrapped_paths)
        assert all(full_suffix(p) == ".unw.tif" for p in paths.unwrapped_paths)
        assert all(p.exists() for p in paths.conncomp_paths)
        assert all(p.exists() for p in paths.timeseries_paths)
        assert all(full_suffix(p) == ".tif" for p in paths.timeseries_paths)

        # check the network size
        assert len(paths.unwrapped_paths) > len(paths.timeseries_paths)


def test_displacement_run_single_official_opera_naming(
    opera_slc_files_official: list[Path],
    # weather_model_files: list[Path],
    # tec_files: list[Path],
    # dem_file: Path,
    # opera_static_files_official: list[Path],
    tmpdir,
):
    with tmpdir.as_cwd():
        cfg = config.DisplacementWorkflow(
            cslc_file_list=opera_slc_files_official,
            input_options={"subdataset": "/data/VV"},
            output_options={"strides": {"x": 2, "y": 2}},
            interferogram_network={"max_bandwidth": 1},
            phase_linking={
                "ministack_size": 500,
            },
            # # TODO: this is not working
            # # either move to disp-s1 test with real data,
            # # or.. something else
            # correction_options={
            #     "troposphere_files": weather_model_files,
            #     "ionosphere_files": tec_files,
            #     "dem_file": dem_file,
            #     "geometry_files": opera_static_files_official,
            # },
            unwrap_options={"run_unwrap": True},
        )
        paths = displacement.run(cfg)
        assert paths.timeseries_paths is not None
        assert all(p.exists() for p in paths.timeseries_paths)
        assert all(full_suffix(p) == ".tif" for p in paths.timeseries_paths)


def run_displacement_stack(
    path, file_list: list[Path], run_unwrap: bool = False, ministack_size: int = 500
):
    cfg = config.DisplacementWorkflow(
        cslc_file_list=file_list,
        input_options={"subdataset": "/data/VV"},
        work_directory=path,
        phase_linking={
            "ministack_size": ministack_size,
        },
        unwrap_options={"run_unwrap": run_unwrap},
        log_file=Path() / "dolphin.log",
    )
    paths = displacement.run(cfg)
    if run_unwrap:
        assert paths.timeseries_paths is not None
        assert all(full_suffix(p) == ".tif" for p in paths.timeseries_paths)


def test_stack_with_compSLCs(opera_slc_files, tmpdir):
    with tmpdir.as_cwd():
        p1 = Path("first_run")
        run_displacement_stack(p1, opera_slc_files, run_unwrap=True)
        # Find the compressed SLC files
        new_comp_slcs = sorted(p1.rglob("compressed_*"))

        p2 = Path("second_run")
        # Add the first compressed SLC in place of first real one and run again
        by_burst = group_by_burst(opera_slc_files)
        new_real_slcs = list(flatten(v[1:] for v in by_burst.values()))
        new_file_list = new_comp_slcs + new_real_slcs

        run_displacement_stack(p2, new_file_list)

        # Now the results should be the same (for the file names)
        # check the ifg folders
        ifgs1 = sorted((p1 / "interferograms").glob("*.int.tif"))
        ifgs2 = sorted((p2 / "interferograms").glob("*.int.tif"))
        assert len(ifgs1) > 0
        assert [f.name for f in ifgs1] == [f.name for f in ifgs2]


def test_separate_workflow_runs(slc_file_list, tmp_path):
    """Check that manually running the workflow results in the same
    interferograms as one sequential run.
    """
    p_all = tmp_path / "all"
    run_displacement_stack(p_all, slc_file_list, ministack_size=10)
    all_ifgs = sorted((p_all / "interferograms").glob("*.int.tif"))
    assert len(all_ifgs) == 29

    p1 = tmp_path / Path("first")
    ms = 10
    # Split into batches of 10
    file_batches = [slc_file_list[i : i + ms] for i in range(0, len(slc_file_list), ms)]
    assert len(file_batches) == 3
    assert all(len(b) == 10 for b in file_batches)
    run_displacement_stack(p1, file_batches[0])
    new_comp_slcs1 = sorted((p1 / "linked_phase").glob("compressed_*"))
    assert len(new_comp_slcs1) == 1
    ifgs1 = sorted((p1 / "interferograms").glob("*.int.tif"))
    assert len(ifgs1) == 9

    p2 = tmp_path / Path("second")
    files2 = new_comp_slcs1 + file_batches[1]
    run_displacement_stack(p2, files2)
    new_comp_slcs2 = sorted((p2 / "linked_phase").glob("compressed_*"))
    assert len(new_comp_slcs2) == 1
    ifgs2 = sorted((p2 / "interferograms").glob("*.int.tif"))
    assert len(ifgs2) == 10

    p3 = tmp_path / Path("third")
    files3 = new_comp_slcs1 + new_comp_slcs2 + file_batches[2]
    run_displacement_stack(p3, files3)
    ifgs3 = sorted((p3 / "interferograms").glob("*.int.tif"))
    assert len(ifgs3) == 10

    all_ifgs_names = [f.name for f in all_ifgs]
    batched_names = [f.name for f in ifgs1 + ifgs2 + ifgs3]
    assert all_ifgs_names == batched_names

    # Last, try one where we dont have the first CCSLC
    # The metadata should still tell it what the reference date is,
    # So the outputs should be the same
    p3_b = tmp_path / Path("third")
    files3_b = new_comp_slcs2 + file_batches[2]
    run_displacement_stack(p3_b, files3_b)
    ifgs3_b = sorted((p3_b / "interferograms").glob("*.int.tif"))
    assert len(ifgs3_b) == 10
    # Names should be the same as the previous run
    assert [f.name for f in ifgs3_b] == [f.name for f in ifgs3]



from __future__ import annotations

import logging
from pathlib import Path
from typing import Literal

from pydantic import (
    BaseModel,
    Field,
    PrivateAttr,
    field_validator,
)

from ._enums import UnwrapMethod

logger = logging.getLogger(__name__)

__all__ = [
    "UnwrapOptions",
    "PreprocessOptions",
    "SnaphuOptions",
    "TophuOptions",
    "SpurtOptions",
]


def to_tuple(v: int | tuple[int, int] | None) -> tuple[int, int]:
    """Convert the input value to a tuple of two integers.

    Parameters
    ----------
    v : Union[int, Tuple[int, int], None]
        The value to be converted.

    Returns
    -------
    Tuple[int, int]
        Tuple containing two integers.

    Examples
    --------
    >>> to_tuple(3)
    (3, 3)
    >>> to_tuple((2, 4))
    (2, 4)
    >>> to_tuple(None)
    (1, 1)

    """
    if v is None:
        return (1, 1)
    if isinstance(v, int):
        return (v, v)
    return v


class PreprocessOptions(BaseModel, extra="forbid"):
    alpha: float = Field(
        0.5,
        description="Adaptive phase (Goldstein) filter exponent parameter.",
        ge=0.0,
        le=1.0,
    )
    max_radius: int = Field(
        51,
        ge=0.0,
        description="(for interpolation) Maximum radius to find scatterers.",
    )
    interpolation_cor_threshold: float = Field(
        0.5,
        description=(
            "Threshold on the correlation raster to use for interpolation. Pixels with"
            " less than this value are replaced by a weighted combination of"
            " neighboring pixels."
        ),
        ge=0.0,
        le=1.0,
    )


class SnaphuOptions(BaseModel, extra="forbid"):
    ntiles: tuple[int, int] = Field(
        (1, 1),
        description=(
            "Number of tiles to split the inputs into using SNAPHU's internal tiling."
        ),
    )
    tile_overlap: tuple[int, int] = Field(
        (0, 0),
        description=(
            "Amount of tile overlap (in pixels) along the (row, col) directions."
        ),
    )
    n_parallel_tiles: int = Field(
        1,
        description="Number of tiles to unwrap in parallel for each interferogram.",
    )
    init_method: Literal["mcf", "mst"] = Field(
        "mcf",
        description="Initialization method for SNAPHU.",
    )
    cost: Literal["defo", "smooth"] = Field(
        "smooth",
        description="Statistical cost mode method for SNAPHU.",
    )

    _to_tuple = field_validator("ntiles", "tile_overlap", mode="before")(to_tuple)


class TophuOptions(BaseModel, extra="forbid"):
    ntiles: tuple[int, int] = Field(
        (1, 1),
        description="Number of tiles to split the inputs into",
    )
    downsample_factor: tuple[int, int] = Field(
        (1, 1),
        description="Extra multilook factor to use for the coarse unwrap.",
    )
    init_method: Literal["mcf", "mst"] = Field(
        "mcf",
        description="Initialization method for SNAPHU.",
    )
    cost: Literal["defo", "smooth"] = Field(
        "smooth",
        description="Statistical cost mode method for SNAPHU.",
    )

    _to_tuple = field_validator("ntiles", "downsample_factor", mode="before")(to_tuple)


class SpurtGeneralSettings(BaseModel):
    use_tiles: bool = Field(default=True, description="Tile up data spatially.")


class SpurtTilerSettings(BaseModel):
    """Class for holding tile generation settings."""

    max_tiles: int = Field(16, description="Maximum number of tiles allowed.", ge=1)
    target_points_for_generation: int = Field(
        120_000,
        description="Number of points used for determining tiles based on density.",
        gt=0,
    )
    target_points_per_tile: int = Field(
        800_000, description="Target points per tile when generating tiles.", gt=0
    )
    dilation_factor: float = Field(
        0.05,
        description=(
            "Dilation factor of non-overlapping tiles. 0.05 would lead to 5 percent"
            " dilation of the tile."
        ),
        ge=0.0,
    )


class SpurtSolverSettings(BaseModel):
    t_worker_count: int = Field(
        default=1,
        alias="t_worker_count",
        description=(
            "Number of workers for temporal unwrapping in parallel. Set value to <=0 to"
            " let workflow use default workers (ncpus - 1)."
        ),
    )
    s_worker_count: int = Field(
        default=1,
        description=(
            "Number of workers for spatial unwrapping in parallel. Set value to <=0"
            " to let workflow use (ncpus - 1)."
        ),
    )
    links_per_batch: int = Field(
        default=50_000,
        description=(
            "Temporal unwrapping operations over spatial links are performed in batches"
            " and each batch is solved in parallel."
        ),
        gt=0,
    )
    t_cost_type: Literal["constant", "distance", "centroid"] = Field(
        default="constant",
        description="Temporal unwrapping costs.",
    )
    t_cost_scale: float = Field(
        default=100.0,
        description="Scale factor used to compute edge costs for temporal unwrapping.",
        gt=0.0,
    )
    s_cost_type: Literal["constant", "distance", "centroid"] = Field(
        default="constant", description="Spatial unwrapping costs.", alias="s_cost_type"
    )
    s_cost_scale: float = Field(
        default=100.0,
        description="Scale factor used to compute edge costs for spatial unwrapping.",
        gt=0.0,
    )


class SpurtMergerSettings(BaseModel):
    min_overlap_points: int = Field(
        default=25,
        description="Minimum number of overlap pixels to be considered valid.",
    )
    method: Literal["dirichlet"] = Field(
        default="dirichlet", description="Currently, only 'dirichlet' is supported."
    )
    bulk_method: Literal["integer", "L2"] = Field(
        default="L2", description="Method used to estimate bulk offset between tiles."
    )


class SpurtOptions(BaseModel, extra="forbid"):
    """Options for running 3D unwrapping on a set of interferograms.

    Uses [`spurt`](https://github.com/isce-framework/spurt) to run the
    temporal/spatial unwrapping. Options are passed through to `spurt`
    library.

    """

    temporal_coherence_threshold: float = Field(
        0.6,
        description="Temporal coherence to pick pixels used on an irregular grid",
        ge=0.0,
        lt=1.0,
    )
    general_settings: SpurtGeneralSettings = Field(default_factory=SpurtGeneralSettings)
    tiler_settings: SpurtTilerSettings = Field(default_factory=SpurtTilerSettings)
    solver_settings: SpurtSolverSettings = Field(default_factory=SpurtSolverSettings)
    merger_settings: SpurtMergerSettings = Field(default_factory=SpurtMergerSettings)


class UnwrapOptions(BaseModel, extra="forbid"):
    """Options for unwrapping after wrapped phase estimation."""

    run_unwrap: bool = Field(
        True,
        description=(
            "Whether to run the unwrapping step after wrapped phase estimation."
        ),
    )
    run_goldstein: bool = Field(
        False,
        description="Whether to run Goldstein filtering step on wrapped interferogram.",
    )
    run_interpolation: bool = Field(
        False,
        description="Whether to run interpolation step on wrapped interferogram.",
    )
    _directory: Path = PrivateAttr(Path("unwrapped"))
    unwrap_method: UnwrapMethod = UnwrapMethod.SNAPHU
    n_parallel_jobs: int = Field(
        1, description="Number of interferograms to unwrap in parallel."
    )
    zero_where_masked: bool = Field(
        False,
        description=(
            "Set wrapped phase/correlation to 0 where mask is 0 before unwrapping. "
        ),
    )
    preprocess_options: PreprocessOptions = Field(default_factory=PreprocessOptions)
    snaphu_options: SnaphuOptions = Field(default_factory=SnaphuOptions)
    tophu_options: TophuOptions = Field(default_factory=TophuOptions)
    spurt_options: SpurtOptions = Field(default_factory=SpurtOptions)



from __future__ import annotations

import logging
from os import fspath
from pathlib import Path

import numpy as np

from dolphin import io
from dolphin._types import Filename
from dolphin.utils import full_suffix
from dolphin.workflows import UnwrapMethod

from ._constants import CONNCOMP_SUFFIX, DEFAULT_CCL_NODATA
from ._utils import _redirect_unwrapping_log, _zero_from_mask

logger = logging.getLogger(__name__)

__all__ = ["unwrap_isce3"]


def unwrap_isce3(
    ifg_filename: Filename,
    corr_filename: Filename,
    unw_filename: Filename,
    mask_file: Filename | None = None,
    unwrap_method: UnwrapMethod = UnwrapMethod.PHASS,
    ccl_nodata: int = DEFAULT_CCL_NODATA,
    zero_where_masked: bool = False,
) -> tuple[Path, Path]:
    """Unwrap a single interferogram using isce3 or tophu.

    Parameters
    ----------
    ifg_filename : Filename
        Path to input interferogram.
    corr_filename : Filename
        Path to input correlation file.
    unw_filename : Filename
        Path to output unwrapped phase file.
    mask_file : Filename, optional
        Path to binary byte mask file, by default None.
        Assumes that 1s are valid pixels and 0s are invalid.
    zero_where_masked : bool, optional
        Set wrapped phase/correlation to 0 where mask is 0 before unwrapping.
        If not mask is provided, this is ignored.
        By default True.
    ccl_nodata : int, default = 65535
        Nodata value to use in output connected component raster
    unwrap_method : UnwrapMethod or str, optional, default = "phass"
        Choice of unwrapping algorithm to use.
        Choices: {"icu", "phass"} (snaphu is done by snaphu-py)

    Returns
    -------
    unw_path : Path
        Path to output unwrapped phase file.
    conncomp_path : Path
        Path to output connected component label file.

    """
    # Now we're using PHASS or ICU within isce3
    from isce3.io import Raster
    from isce3.unwrap import ICU, Phass

    shape = io.get_raster_xysize(ifg_filename)[::-1]
    corr_shape = io.get_raster_xysize(corr_filename)[::-1]
    if shape != corr_shape:
        msg = f"correlation {corr_shape} and interferogram {shape} shapes don't match"
        raise ValueError(msg)

    ifg_raster = Raster(fspath(ifg_filename))
    corr_raster = Raster(fspath(corr_filename))
    UNW_SUFFIX = full_suffix(unw_filename)

    # Get the driver based on the output file extension
    if Path(unw_filename).suffix == ".tif":
        driver = "GTiff"
        opts = list(io.DEFAULT_TIFF_OPTIONS)
    else:
        driver = "ENVI"
        opts = list(io.DEFAULT_ENVI_OPTIONS)

    unw_nodata = -10_000 if unwrap_method == UnwrapMethod.PHASS else 0
    # Create output rasters for unwrapped phase & connected component labels.
    # Writing with `io.write_arr` because isce3 doesn't have creation options
    io.write_arr(
        arr=None,
        output_name=unw_filename,
        driver=driver,
        like_filename=ifg_filename,
        dtype=np.float32,
        nodata=unw_nodata,
        options=opts,
    )
    conncomp_filename = str(unw_filename).replace(UNW_SUFFIX, CONNCOMP_SUFFIX)
    io.write_arr(
        arr=None,
        output_name=conncomp_filename,
        driver="GTiff",
        dtype=np.uint16,
        like_filename=ifg_filename,
        nodata=ccl_nodata,
        options=io.DEFAULT_TIFF_OPTIONS,
    )

    # The different raster classes have different APIs, so we need to
    # create the raster objects differently.
    unw_raster = Raster(fspath(unw_filename), True)
    conncomp_raster = Raster(fspath(conncomp_filename), True)

    _redirect_unwrapping_log(unw_filename, unwrap_method.value)

    if zero_where_masked and mask_file is not None:
        logger.info(f"Zeroing phase/corr of pixels masked in {mask_file}")
        zeroed_ifg_file, zeroed_corr_file = _zero_from_mask(
            ifg_filename, corr_filename, mask_file
        )
        corr_raster = Raster(fspath(zeroed_corr_file))
        ifg_raster = Raster(fspath(zeroed_ifg_file))

    logger.info(
        f"Unwrapping size {(ifg_raster.length, ifg_raster.width)} {ifg_filename} to"
        f" {unw_filename} using {unwrap_method.value}"
    )
    if unwrap_method == UnwrapMethod.PHASS:
        # TODO: expose the configuration for phass?
        # If we ever find cases where changing help, then yes we should
        # coherence_thresh: float = 0.2
        # good_coherence: float = 0.7
        # min_region_size: int = 200
        unwrapper = Phass()
        unwrapper.unwrap(
            ifg_raster,
            corr_raster,
            unw_raster,
            conncomp_raster,
        )
    else:
        unwrapper = ICU(buffer_lines=shape[0])

        unwrapper.unwrap(
            unw_raster,
            conncomp_raster,
            ifg_raster,
            corr_raster,
        )
    if zero_where_masked and mask_file is not None:
        logger.info(f"Zeroing unw/conncomp of pixels masked in {mask_file}")
        return _zero_from_mask(unw_filename, conncomp_filename, mask_file)

    del unw_raster, conncomp_raster
    return Path(unw_filename), Path(conncomp_filename)



import logging

import numba
import numpy as np
from numpy.typing import ArrayLike

from .similarity import get_circle_idxs

logger = logging.getLogger(__name__)


def interpolate(
    ifg: ArrayLike,
    weights: ArrayLike,
    weight_cutoff: float = 0.5,
    num_neighbors: int = 20,
    max_radius: int = 51,
    min_radius: int = 0,
    alpha: float = 0.75,
) -> np.ndarray:
    """Interpolate a complex interferogram based on pixel weights.

    Build upon persistent scatterer interpolation used in
    [@Chen2015PersistentScattererInterpolation] and
    [@Wang2022AccuratePersistentScatterer] by allowing floating-point weights
    instead of 0/1 PS weights.

    Parameters
    ----------
    ifg : np.ndarray, 2D complex array
        wrapped interferogram to interpolate
    weights : 2D float array
        Array of weights from 0 to 1 indicating how strongly to weigh
        the ifg values when interpolating.
        A special case of this is a PS mask where
            weights[i,j] = True if radar pixel (i,j) is a PS
            weights[i,j] = False if radar pixel (i,j) is not a PS
        Can also pass a coherence image to use as weights.
    weight_cutoff: float
        Threshold to use on `weights` so that pixels where
        `weight[i, j] < weight_cutoff` have phase values replaced by
        an interpolated value.
        The default is 0.5: pixels with weight less than 0.5 are replaced with a
        smoothed version of the surrounding pixels.
    num_neighbors: int (optional)
        number of nearest PS pixels used for interpolation
        num_neighbors = 20 by default
    max_radius : int (optional)
        maximum radius (in pixels) for PS searching
        max_radius = 51 by default
    min_radius : int (optional)
        minimum radius (in pixels) for PS searching
        max_radius = 0 by default
    alpha : float (optional)
        hyperparameter controlling the weight of PS in interpolation: smaller
        alpha means more weight is assigned to PS closer to the center pixel.
        alpha = 0.75 by default

    Returns
    -------
    interpolated_ifg : 2D complex array
        interpolated interferogram with the same amplitude, but different
        wrapped phase at non-ps pixels.

    """
    nrow, ncol = weights.shape

    weights_float = np.clip(weights.astype(np.float32), 0, 1)
    # Ensure weights are between 0 and 1
    if np.any(weights_float > 1):
        logger.warning("weights array has values greater than 1. Clipping to 1.")
    if np.any(weights_float < 0):
        logger.warning("weights array has negative values. Clipping to 0.")
    weights_float = np.clip(weights_float, 0, 1)

    interpolated_ifg = np.zeros((nrow, ncol), dtype=np.complex64)

    indices = np.array(
        get_circle_idxs(max_radius, min_radius=min_radius, sort_output=False)
    )

    _interp_loop(
        ifg,
        weights_float,
        weight_cutoff,
        num_neighbors,
        alpha,
        indices,
        interpolated_ifg,
    )
    return interpolated_ifg


@numba.njit(parallel=True)
def _interp_loop(
    ifg, weights, weight_cutoff, num_neighbors, alpha, indices, interpolated_ifg
):
    nrow, ncol = weights.shape
    nindices = len(indices)
    for r0 in numba.prange(nrow):
        for c0 in range(ncol):
            if weights[r0, c0] >= weight_cutoff:
                interpolated_ifg[r0, c0] = ifg[r0, c0]
                continue

            csum = 0.0 + 0j
            counter = 0
            r2 = np.zeros(num_neighbors, dtype=np.float64)
            cphase = np.zeros(num_neighbors, dtype=np.complex128)

            for i in range(nindices):
                idx = indices[i]
                r = r0 + idx[0]
                c = c0 + idx[1]

                if (
                    (r >= 0)
                    and (r < nrow)
                    and (c >= 0)
                    and (c < ncol)
                    and weights[r, c] >= weight_cutoff
                ):
                    # calculate the square distance to the center pixel
                    r2[counter] = idx[0] ** 2 + idx[1] ** 2

                    cphase[counter] = np.exp(1j * np.angle(ifg[r, c]))
                    counter += 1
                    if counter >= num_neighbors:
                        break

            # `counter` got up to one more than the number of elements
            # The last one will be the largest radius
            r2_norm = (r2[counter - 1] ** alpha) / 2
            for i in range(counter):
                csum += np.exp(-r2[i] / r2_norm) * cphase[i]

            interpolated_ifg[r0, c0] = np.abs(ifg[r0, c0]) * np.exp(1j * np.angle(csum))



import argparse
from typing import TYPE_CHECKING, Any, Optional, Sequence

if TYPE_CHECKING:
    _SubparserType = argparse._SubParsersAction[argparse.ArgumentParser]
else:
    _SubparserType = Any


def run(
    config_file: str,
    debug: bool = False,
) -> None:
    """Run the displacement workflow.

    Parameters
    ----------
    config_file : str
        YAML file containing the workflow options.
    debug : bool, optional
        Enable debug logging, by default False.

    """
    # rest of imports here so --help doesn't take forever

    from . import displacement
    from .config import DisplacementWorkflow

    cfg = DisplacementWorkflow.from_yaml(config_file)
    displacement.run(cfg, debug=debug)


def get_parser(
    subparser: Optional[_SubparserType] = None, subcommand_name: str = "run"
) -> argparse.ArgumentParser:
    """Set up the command line interface."""
    metadata = {
        "description": "Run a displacement workflow",
        "formatter_class": argparse.ArgumentDefaultsHelpFormatter,
    }
    if subparser:
        # Used by the subparser to make a nested command line interface
        parser = subparser.add_parser(subcommand_name, **metadata)  # type: ignore[arg-type]

    else:
        parser = argparse.ArgumentParser(**metadata)  # type: ignore[arg-type]

    parser.add_argument(
        "config_file",
        help="Name of YAML configuration file describing workflow options.",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Print debug messages to the log.",
    )
    parser.set_defaults(run_func=run)
    return parser


def main(args: Optional[Sequence[str]] = None) -> None:
    """Get the command line arguments and run the workflow."""
    parser = get_parser()
    parsed_args = parser.parse_args(args)

    run(parsed_args.config_file, debug=parsed_args.debug)


if __name__ == "__main__":
    main()
